from pathlib import Path
from typing import *

import pandas


def read_edges_filename(path:Path)->pandas.DataFrame:
	# Use a dataframe in case the edges file is an excel file rather than a csv file.
	if path.suffix in ['.xlsx', '.xls']:
		table = pandas.read_excel(path)
	elif path.suffix == '.csv':
		table = pandas.read_csv(path)
	elif path.suffix == '.tsv':
		table = pandas.read_table(path)
	else:
		message = f"The given filetype is not supported. Expected one of ['.csv', '.tsv', '.xls', '.xlsx'], received '{path.suffix}'"
		raise ValueError(message)
	return table

def main(edgesio:Union[Path, pandas.Series], annotations:Dict[str,str] = None, palette:Dict[str,str] = None, filename = None):
	"""
		Generates a lineageplot of the given data.
		Parameters
		----------
		edgesio: Union[Path, pandas.Series]
			The lineage for this workflow. If `edgesio` is a file it may contain additional information via these columns:
			- 'identity': The name of the genotype
			- `parent`: The name of the genotype's most curent ancestor
			- `annotation': An annotation to include alongside the genotype
			- `color`: The color to use when drawing the lineageplot
		annotations: Dict[str,str]
			An optional dictionary mapping annotations to the genotypes in `edges`. Overrides any annotations found in the `edges` file.
		palette: Dict[str,str]
			A colorscheme mapping genotype names to colors. May be given as a file, or extracted from the `edges` file.
		filename: Path
			The file to save the lineageplot as.
	"""

	if isinstance(edgesio, Path):
		table_edges = read_edges_filename(edgesio)
	else:
		# Assume `edgesio` is a pandas DataFrame
		table_edges = edgesio

	# Convert the table columns to lowercase so we don't have to work with multiple capitalizations.
	# Also convert to string in case other type managed to becom column labels.
	table_edges.columns = [str(i).lower for i in table_edges.columns]
	# Try to get the option from the input DataFrame.
	if annotations is None and 'annotation' in table_edges:
		annotations = table_edges['annotation'].to_dict()

	if palette is None:
		if 'color' in table_edges.column:
			# Use the given color.
			palette = table_edges['color'].to_dict()
		else:
			# Generate the palette
			pass


if __name__ == "__main__":
	pass


"""
	Generates a palette using the annotations available from the input trajectories table.
"""
from typing import Dict, List

from loguru import logger


def generate_annotation_palette(genotype_annotations: Dict[str, List[str]], annotation_palette: Dict[str, str]) -> Dict[str, str]:
	"""

	Parameters
	----------
	genotype_annotations: Dict[str,List[str]]
		The annotations grouped under a given genome.
	annotation_palette: Dict[str,str]
		A map of annotations to a preset color. Genotypes with no annotations will not be assigned a color.

	Returns
	-------

	"""
	palette = dict()
	for genotype_label, genotype_annotation in genotype_annotations.items():
		candidate_colors = [annotation_palette.get(i) for i in genotype_annotation]
		candidate_colors = sorted(set(i for i in candidate_colors if i))
		if len(candidate_colors) > 1:
			logger.warning(f"Found multiple candidate colors for genotype '{genotype_label}': {candidate_colors}")
		try:
			color = candidate_colors[0]
		except IndexError:
			color = None
		palette[genotype_label] = color
	return palette



import subprocess
from pathlib import Path
from typing import Dict, List


def generate_r_script(trajectory: Path, population: Path, edges: Path,
		color_palette: Dict[str, str], genotype_labels: List[str]) -> str:
	# `color_palette` should be an OrderedDict.
	genotype_labels = sorted(genotype_labels)

	script_colors = ",".join(['"{}"'.format(color_palette.get(k, "#333333")) for k in genotype_labels])

	script = """
	library(ggplot2)
	library(ggmuller)

	population <- read.table("{population}", header=TRUE)
	edges <- read.table("{edges}", header=TRUE)

	Muller_df <- get_Muller_df(edges, population)
	palette <- c({palette})

	ggplot(Muller_df, aes_string(x = "Generation", y = "Frequency", group = "Group_id", fill = "Identity", colour = "Identity")) + 
    geom_area() +
    theme(legend.position = "right") +
    guides(linetype = FALSE, color = FALSE) + 
    scale_y_continuous(labels = 25 * (0:4), name = "Percentage", expand=c(0,0)) +
    scale_x_continuous(expand=c(0,0)) + 
    scale_fill_manual(name = "Identity", values = palette) +
    scale_color_manual(values = palette)

	# ggsave("{output}", height = 10, width = 10)
	""".format(
		trajectory = trajectory,
		population = population.absolute(),
		edges = edges.absolute(),
		output = "filename",
		palette = script_colors
	)
	script = '\n'.join(i.strip() for i in script.split('\n'))

	return script


def execute_r_script(path: Path, script: str) -> Path:
	""" Executes `script` by first saving it to `path` then calling `Rscript` to run it."""
	path.write_text(script)
	subprocess.run(
		['Rscript', '--vanilla', '--silent', path],
		stdout = subprocess.PIPE,
		stderr = subprocess.PIPE
	)
	# Running the script will generate an Rplots.pdf file which we don't care about.
	# `ggsave` is used to save the plot instead.
	_extra_file = Path.cwd() / "Rplots.pdf"
	if _extra_file.exists():
		_extra_file.unlink()
	return path



import csv
import itertools
import math
import re
from pathlib import Path
from typing import *

import pandas
from loguru import logger

NumericType = Union[int, float]
IterableValues = Union[List[NumericType], pandas.Series]
NUMERIC_REGEX = re.compile("^.?(?P<number>[\d]+)")

def _coerce_to_series(item:Any)->pandas.Series:
	if not isinstance(item, pandas.Series):
		item = pandas.Series(item)
	return item

def _coerce_to_list(item:Any)->List[Any]:
	if isinstance(item, list):
		# Nothing to change
		result = item
	elif isinstance(item, pandas.Series):
		result = item.tolist()
	else:
		try:
			result = list(item)
		except ValueError:
			message = f"Could not coerce this to a list: '{type(item)}' -> {item}"
			raise ValueError(message)
	return result
def checkdir(path: Union[str, Path]) -> Path:
	path = Path(path)
	if not path.exists():
		path.mkdir()
	return path


def get_numeric_columns(columns: List[Union[str, int, float]]) -> List[Union[str, int, float]]:
	numeric_columns = list()
	for column in columns:
		if isinstance(column, str):
			match = NUMERIC_REGEX.search(column)
			if match:
				col = match.groupdict()['number']
			else:
				continue
		else:
			col = column

		try:
			int(col)
		except (ValueError, TypeError):
			continue
		numeric_columns.append(column)
	return numeric_columns

def get_numeric_table(table:pandas.DataFrame)->pandas.DataFrame:
	""" Returns a version of the input table with only the columns related to timepoints.
		The columns will also be converted to numeric values.
	"""
	table = table[get_numeric_columns(table.column)]
	table.columns = [int(i) for i in table.columns]

	return table

def map_trajectories_to_genotype(genotype_members: pandas.Series) -> Dict[str, str]:
	""" Maps each trajectory to the genotype it belongs to."""
	trajectory_to_genotype = dict()
	for genotype_label, members in genotype_members.items():
		for member in members.split('|'):
			trajectory_to_genotype[member] = genotype_label
	return trajectory_to_genotype


def get_valid_points(left_trajectory: pandas.Series, right_trajectory: pandas.Series, dlimit: float, flimit: Optional[float] = None,
		inner: bool = False) -> Tuple[pandas.Series, pandas.Series]:
	"""
		Filters out timepoints that do not satisfy the detection criteria.
	Parameters
	----------
	left_trajectory: pandas.Series
	right_trajectory: pandas.Series
	dlimit: float
		Removes the timepoint if both points do no exceed this value.
	flimit: float
		If given, all points exceeding this value are treated as "undetected".
	inner: bool; default False
		If `true`, both points must exceed the detection limit for a given timepoint to be considered valid.

	Returns
	-------

	"""

	if flimit is not None:
		# Mask the values so they are considered below the detection limit.
		# Assign a value of -1 so that they will be excluded even if the detection limit is 0.
		# list comprehension is ~6X faster than mask()
		left = pandas.Series([(-1 if s > 0.97 else s) for s in left_trajectory.values], index = left_trajectory.index)
		right = pandas.Series([(-1 if s > 0.97 else s) for s in right_trajectory.values], index = right_trajectory.index)
	else:
		left, right = left_trajectory, right_trajectory

	if inner:
		# list comprehensions are ~10X faster than using the built-in pandas methods.
		at_least_one_detected = [(l > dlimit and r > dlimit) for l, r in zip(left.values, right.values)]
	else:
		at_least_one_detected = [(l > dlimit or r > dlimit) for l, r in zip(left.values, right.values)]
	at_least_one_detected = pandas.Series(at_least_one_detected, index = left.index)

	# Remove indicies where the series value falls below the detection limit. This should include the masked fixed values.
	at_least_one_detected_reduced = at_least_one_detected[at_least_one_detected]
	if at_least_one_detected_reduced.empty:
		# There are no shared timepoints between the series. Assign index_min and index_max to the same number, which will result in an empty dataframe.
		position_index_min = position_index_max = 0
	else:
		# Apparently the min() and max functions now work with strings as well as numbers.
		# Cast the numbers to float so the typeerror is thrown correctly.
		try:
			position_index_min_value = min(at_least_one_detected_reduced.index, key = lambda s: float(s))
			position_index_max_value = max(at_least_one_detected_reduced.index, key = lambda s: float(s))
		except TypeError:
			# The indicies are str and we can't use min() or max(). Assume the indicies are already sorted.
			position_index_min_value = at_least_one_detected_reduced.index[0]
			position_index_max_value = at_least_one_detected_reduced.index[-1]

		position_index_min = left.index.get_loc(position_index_min_value)
		position_index_max = left.index.get_loc(position_index_max_value)
		# Since we want to include the last index, increment position_index_max by one.
		position_index_max += 1
	result_left = left_trajectory[position_index_min:position_index_max]
	result_right = right_trajectory[position_index_min:position_index_max]

	return result_left, result_right


get_detected_points = get_valid_points


def calculate_luminance(color: str) -> float:
	# 0.299 * color.R + 0.587 * color.G + 0.114 * color.B
	red = int(color[1:3], 16)
	green = int(color[3:5], 16)
	blue = int(color[5:], 16)

	lum = (.299 * red) + (.587 * green) + (.114 * blue)
	return lum / 255


def format_inconsistency_matrix(inconsistency_matrix) -> pandas.DataFrame:
	inconsistency_table = pandas.DataFrame(inconsistency_matrix, columns = ['mean', 'std', 'observations', 'statistic'])
	inconsistency_table['observations'] = inconsistency_table['observations'].astype(int)
	return inconsistency_table


# noinspection PyTypeChecker,PyUnresolvedReferences
def fixed_immediately(trajectory: pandas.Series, dlimit: float, flimit: float) -> bool:
	mask1 = (trajectory <= dlimit)
	mask2 = (trajectory >= flimit)
	return (mask1.sum() + mask2.sum()) == len(trajectory)


def fixed(trajectory: pandas.Series, flimit: float) -> bool:
	""" Tests whether the input series fixed at any point.
		If it throws a ValueError due to being an array, check the input data for duplicate labels
	"""
	try:
		return any(i > flimit for i in trajectory.values)
	except ValueError:
		message = "Encountered an error when selecting 'fixed' trajectories. This is usually caused by duplicate trajectory ids in the input table."
		if isinstance(trajectory, pandas.DataFrame):
			message += f"\nRename these trajectories: {list(trajectory.index)}"
		raise ValueError(message)
	except TypeError as exception:
		logger.warning(f"Trajectory: {trajectory.values}")
		logger.warning(f"flimit: {flimit}")

		raise exception

def get_fixed(trajectory:IterableValues, flimit)->pandas.Series:
	""" Returns a pandas.Series object with only the timepoints where the series was fixed."""
	# Usa a dummy value for the `flimit` parameter.
	result = get_intermediate(trajectory, dlimit = flimit, flimit = 2)
	return result
def get_undetected(trajectory:IterableValues, dlimit:float)->pandas.Series:
	return get_intermediate(trajectory, dlimit = -2, flimit = dlimit)
def get_intermediate(trajectory:IterableValues, dlimit:float, flimit:float)->pandas.Series:
	""" Tests whether the input series had timepoints that were neither undetected nor fixed.
		Returns a pandas.Series object with the indecies and values that satisfy this criteria.
	"""
	trajectory = _coerce_to_series(trajectory)
	is_intermediate = trajectory.apply(lambda s: dlimit <= s <= flimit)
	# remove the timepoints which are undetected or fixed
	result = trajectory[is_intermediate]
	return result



def only_fixed(trajectory: pandas.Series, dlimit: float, flimit: float) -> bool:
	""" Tests whether the series immediately fixed and stayed fixed."""
	series = ((i > flimit or i < dlimit) for i in trajectory.values)
	return all(series)

def get_first_fixed_timepoint(elements:pandas.Series, flimit:float)->Any:
	""" Returns the first index that was `fixed` """
	elements = _coerce_to_series(elements)
	result = elements.apply(lambda s: s>flimit)
	# Check if any timepoints were fixed
	if result.sum() == 0:
		result = None
	else:
		result = result.idxmax()

	return result

def get_overlap_regions(left:IterableValues, right:IterableValues, flimit:float)->pandas.Series:
	"""
		Returns the regions between `left` and `right` that have overlapping "fixed" values.
		This will be a pandas.Series object with timepoints as the index and boolean values indicating
		if both series were fixed at that timepoint.
	"""
	# Convert to a pandas.Series object if they aren't already
	left = _coerce_to_series(left)
	right = _coerce_to_series(right)

	fixed_left = left.apply(lambda s: s>= flimit)
	fixed_right = right.apply(lambda s: s>= flimit)
	overlapping_regions = fixed_left & fixed_right
	return overlapping_regions

def overlap(left: Union[List[float], pandas.Series], right: Union[List[float], pandas.Series], dlimit: float) -> int:
	result = [(i > dlimit and j > dlimit) for i, j in zip(left.values, right.values)]

	return sum(result)


# noinspection PyTypeChecker
def find_boundaries_fixed(trajectory: pandas.Series, flimit: float) -> Optional[Tuple[int, int]]:
	fixed_series = trajectory[trajectory > flimit]
	# Assume index is sorted to avoid the situation where the index is str.
	if fixed_series.empty: return None
	else:
		return fixed_series.index[0], fixed_series.index[-1]


def _get_git_log() -> str:
	filename = Path(__file__).parent.parent / ".git" / "logs" / "HEAD"
	try:
		contents = filename.read_text()
	except FileNotFoundError:
		contents = ""
	return contents


def get_commit_hash() -> str:
	commit_hash = "n/a"
	contents = _get_git_log()
	if contents:
		contents = contents.split('\n')
		contents = [i.strip() for i in contents if i.strip()]
		reader = csv.reader(contents, delimiter = '\t')
		for line in reader:
			if line:
				hash_string = line[0]
				try:
					commit_hash = hash_string.split()[1]
				except IndexError:
					continue
		commit_hash = commit_hash[:7]
	else:
		commit_hash = "not available"
	return commit_hash


def get_pair_combinations(elements: Iterable[str]) -> Generator[Tuple[str, str], None, None]:
	# Use a generator to avoid using too much memory for large datasets.
	# The only reason to use a list was so tqdm could automatically figure out how many operations had to be done,
	# but this can be calculated manually by passing the total number of trajectories as an additional parameter.
	return itertools.combinations(elements, 2)


def calculate_number_of_combinations(n: int, r:int = 2) -> int:
	""" Calculates the number of pairwise combinations for an iterable of size `elements` using the
		formula `n!/((n-r)!r!)
		Since we only want combinations of pairs (`r` = 2), this can be simplified to
		`n * (n-1).
	"""
	n_factorial = math.factorial(n)
	r_factorial = math.factorial(r)
	n_minus_r_factorial = math.factorial(n-r)
	#return int(n * (n - r)/r)

	result = n_factorial / (n_minus_r_factorial * r_factorial)
	return int(result)

def validate_table_genotypes(table:pandas.DataFrame):
	""" Throws an error if the genotype table does not conform to the expected format.
		This is mainly for debugging while trying to make sure the genotype table is cosistent at every stage.

		Genotype Table Schema:
		- Columns will be in two forms: integers indicating time points, and a str label for the `members` column.
	"""
	assert 'members' in table.columns
	assert all(isinstance(i, int) for i in table.columns if i != 'members')





def validate_series_edges(edges:pandas.Series):
	""" Validates that the edges table is a pandas.Series object and has correct labels.
		This is ment for debugging while trying to make the edges variable more consistent accross scripts.
	"""

	assert isinstance(edges, pandas.Series)
	assert edges.index.name == 'Identity'
	assert edges.name == 'Parent'



from . import projectdata
from .generate_scripts import execute_r_script, generate_r_script
from .generate_tables import GGMuller, generate_trajectory_table
from .genotypecollection import Genotype, GenotypeCollection
from .import_file import *
from .import_tables import import_table
from .import_timeseries import parse_genotype_table, parse_trajectory_table
from .mullerformat import GenerateMullerDataFrame




from pathlib import Path
from typing import Dict, List

import pandas

from muller import graphics
from muller.graphics import Palette

# The filenames should be mapped to their corresponding palette.
# The dictionary maps palette names to the filenames that should be generated.
GraphicPath = Dict[str, List[Path]]


class OutputGeneratorGraphics:
	""" Implemented to avoid having a `god object` controlling all output. It also allows us to decouple the logic
		used to generate the graphics from the logic used to save the data tables.

		This class is designed to be used in parts rather than as a whole so that any workflow (suck as the `ggmuller` workflow)
		can use only the methods it needs.

		There are only two variables needed for multiple processes: the palettes and whether to generate renders of data as well.


	"""

	# TODO: Need to rafactor this so it doesn't need the sample basename.
	def __init__(self, project_folder: Path, palettes: List[Palette], sample_basename: str, render: bool = True, draw_outline: bool = True):
		"""
			Use this method to set up some variables used by multiple workflows, such as `render`.
		"""
		# All we need from the basic_data is the base filename, so that should be passed directly.
		self.render = render
		self.extensions = [".png", ".svg"]

		# Coerce the given palettes into a list of Palette objects. These palettes will be visible by
		# many of the implemented methods.
		if isinstance(palettes, Palette):
			# In case only one Palette object was passed and was not already in a list.
			self.palettes = [palettes]
		else:
			self.palettes: List[Palette] = palettes

		# Each of the files will be prefixed with the population name.
		# Use dataio.PathsGraphics to set up default filenames.

		# Set up the individual graphic generators.
		self.generator_panel_timeseries = graphics.TimeseriesPanel(render = self.render)
		self.generator_plot_timeseries = graphics.TimeseriesPlot(render = self.render)
		self.generator_plot_muller = graphics.MullerPlot(outlines = draw_outline, render = self.render)

	##############################################################################################################################################
	# ---------------------------------------------------------- General Utilites ----------------------------------------------------------------
	##############################################################################################################################################

	@staticmethod
	def get_base_filename(data_basic) -> str:
		if data_basic.program_options.name:
			base_filename = data_basic.program_options.name
		else:
			base_filename = data_basic.filename.stem
		if data_basic.program_options['sheetname'] and data_basic.program_options['sheetname'] != 'Sheet1':
			base_filename += '.' + data_basic.program_options['sheetname']

		return base_filename

	##############################################################################################################################################
	# --------------------------------------------------------------- Panels ---------------------------------------------------------------------
	##############################################################################################################################################

	@staticmethod
	def generate_dendrogram(linkage_matrix, distance_matrix, filename: Path) -> Path:
		# Only need the distance matrix fpr the labels
		labels = distance_matrix.squareform().index
		graphics.plot_dendrogram(linkage_matrix, labels, filename)
		return filename
	@staticmethod
	def generate_heatmap(squareform: pandas.DataFrame, filename: Path) -> Path:

		graphics.plot_heatmap(squareform, filename)
		return filename



from typing import Dict, List, Optional, Tuple

import pandas
from loguru import logger

try:
	from muller.inheritance import scoring
	from muller.inheritance.genotype_ancestry import Ancestry
	from muller import widgets, dataio
except ModuleNotFoundError:
	from . import scoring
	from .genotype_ancestry import Ancestry
	from .. import widgets, dataio


class LineageWorkflow:
	"""
		Orders genotypes by which background they belong to.
	Parameters
	----------
	dlimit: float
		The detection limit
	flimit: float
		The cutoff value to consider a genotype "fixed"
	pvalue: float
		The pvalue to use for statistical tests.
	"""

	def __init__(self, dlimit: float, flimit: float, pvalue: float, weights = (1, 1, 2, 2), conservative:bool = False,debug: bool = False):
		self.dlimit = dlimit
		self.flimit = flimit
		self.pvalue = pvalue
		self.debug = debug
		self.genotype_nests: Optional[Ancestry] = None
		self.conservative = conservative
		self.scorer = scoring.Score(self.dlimit, self.flimit, self.pvalue, weights)

	def __repr__(self)->str:
		return f"LineageWorkflow(dlimit = {self.dlimit}, flimit = {self.flimit}, pvalue = {self.pvalue})"

	def add_known_lineages(self, known_ancestry: Dict[str, str]):
		for identity, parent in known_ancestry.items():
			# TODO need to add a way to prevent circular ancestry links when a user manually assigns ancestry. Current workaround forces the manual parent to be in the root background.
			logger.debug(f"Adding {parent} as a potential background for {identity}")
			self.genotype_nests.add_genotype_to_background(parent, 'genotype-0', priority = 100)
			# Use a dummy priority so that it is selected before other backgrounds.
			self.genotype_nests.add_genotype_to_background(identity, parent, priority = 100)

	def show_ancestry(self, sorted_genotypes: pandas.DataFrame):
		for genotype_label in sorted_genotypes.index:
			candidate = self.genotype_nests.get_highest_priority(genotype_label)
			if self.debug:
				logger.debug(f"{genotype_label}\t{candidate}")

	def run(self, sorted_genotypes: pandas.DataFrame, known_ancestry: Dict[str, str] = None) -> dataio.projectdata.DataGenotypeLineage:
		"""
			Infers the lineage from the given genotype table.
		Parameters
		----------
		sorted_genotypes: pandas.DataFrame
			A dataframe of sorted genotypes based on when the genotype was first detected and first fixed.
		known_ancestry: Dict[str,str]
			Manually-assigned ancestry values. For now, the parent genotype is automatically assigned to the root genotype to prevent
			circular links from forming.
		"""

		initial_background = sorted_genotypes.iloc[0]
		self.genotype_nests = Ancestry(initial_background, timepoints = sorted_genotypes, cautious = self.conservative)
		self.add_known_lineages(known_ancestry if known_ancestry else dict())

		score_records: List[Dict[str, float]] = list()  # Keeps track of the individual score values for each pair

		for unnested_label, unnested_trajectory in sorted_genotypes[1:].iterrows():
			# Iterate over the rest of the table in reverse order. Basically, we start with the newest nest and iterate until we find a nest that satisfies the filters.
			test_table = sorted_genotypes[:unnested_label].iloc[::-1]
			for nested_label, nested_genotype in test_table.iterrows():
				if nested_label == unnested_label: continue
				score_data = self.scorer.score_pair(nested_genotype, unnested_trajectory)
				score_records.append(score_data)
				self.genotype_nests.add_genotype_to_background(unnested_label, nested_label, score_data['totalScore'])

		self.show_ancestry(sorted_genotypes)

		output_data = dataio.projectdata.DataGenotypeLineage(
			table_scores = pandas.DataFrame(score_records),
			clusters = self.genotype_nests, # Used to extract the `edges` table.
		)

		return output_data



def get_maximum_genotype_delta(genotype_deltas: List[Tuple[str, float]]) -> Tuple[str, float]:
	if genotype_deltas:
		correlated_label, correlated_delta = max(genotype_deltas, key = lambda s: s[1])
	else:
		correlated_label = "N/A"  # Shouldn't end up being used.
		correlated_delta = 0
	return correlated_label, correlated_delta



import itertools
from typing import *

import numpy
import pandas
from loguru import logger
from scipy.cluster import hierarchy
from scipy.spatial import distance

try:
	from muller.clustering.metrics.distance_cache import DistanceCache
	from muller.dataio import projectdata
except ModuleNotFoundError:
	from muller.clustering.metrics.distance_cache import DistanceCache
	from muller.dataio import projectdata


def format_linkage_matrix(linkage_table, total_members: Optional[int]) -> pandas.DataFrame:
	linkage_dataframe = pandas.DataFrame(linkage_table, columns = ["left", "right", "distance", "observations"])

	linkage_dataframe['left'] = linkage_dataframe['left'].astype(int)
	linkage_dataframe['right'] = linkage_dataframe['right'].astype(int)
	linkage_dataframe['observations'] = linkage_dataframe['observations'].astype(int)
	if total_members:
		linkage_dataframe['clusterId'] = list(total_members + i for i in range(len(linkage_dataframe)))
	# Rearrange the columns to be more readable.
	linkage_dataframe = linkage_dataframe[['left', 'right', 'distance', 'observations', 'clusterId']]
	return linkage_dataframe


class HierarchalCluster:
	def __init__(self, linkage: str = 'ward', cluster: str = 'distance'):
		self.linkage_method = linkage
		self.cluster_method = cluster

	@staticmethod
	def _add_starting_genotypes(pair_array, starting_genotypes) -> Dict[str, str]:
		for genotype in starting_genotypes:
			combinations = itertools.combinations(genotype, 2)
			values = dict()
			for a, b in combinations:
				values[a, b] = 0
				values[b, a] = 0
			pair_array = pair_array.update(values)
		return pair_array

	@staticmethod
	def _cluster_by_distance(linkage_table: numpy.ndarray, cutoff: float) -> numpy.ndarray:
		""" Try to infer a good distance cutoff by detecting the first changepoint in the sorted array of distances."""
		clusters = hierarchy.fcluster(linkage_table, t = cutoff, criterion = 'distance')
		return clusters

	@staticmethod
	def _cluster_by_monocrit(linkage_table: numpy.ndarray, cutoff: float, inconsistent: pandas.DataFrame) -> numpy.ndarray:
		MR = hierarchy.maxRstat(linkage_table, inconsistent.values, 1)
		clusters = hierarchy.fcluster(linkage_table, t = cutoff, criterion = 'monocrit', monocrit = MR)
		return clusters

	@staticmethod
	def _cluster_by_inconsistent(linkage_table: numpy.ndarray, cutoff: float, inconsistent: pandas.DataFrame) -> numpy.ndarray:
		clusters = hierarchy.fcluster(linkage_table, t = cutoff, criterion = 'inconsistent', R = inconsistent.values)
		return clusters

	@staticmethod
	def _get_inconsistent(distances: numpy.ndarray) -> pandas.DataFrame:
		inconsistent = hierarchy.inconsistent(distances, 10)
		inc = pandas.DataFrame(inconsistent)
		inc.columns = ["mean", "std", "count", "coefficient"]
		return inc

	@staticmethod
	def _label_clusters(clusters: numpy.ndarray, labels: Iterable[str]) -> List[List[str]]:
		""" Maps the original labels to the corresponding numerical index in `clusters`"""
		cluster_map = {i: list() for i in clusters}
		for i, j in zip(clusters, labels):
			cluster_map[i].append(j)
		return list(cluster_map.values())


	def link_clusters(self, distances: numpy.ndarray, num: int) -> pandas.DataFrame:
		Z = hierarchy.linkage(distances, method = self.linkage_method, optimal_ordering = True)

		return format_linkage_matrix(Z, num)

	def cluster(self, linkage_table: pandas.DataFrame, cutoff: float, labels: Optional[Iterable[str]] = None) -> List[List[Union[int, str]]]:
		inconsistent = self._get_inconsistent(linkage_table.values)
		if self.cluster_method == 'distance':
			clusters = self._cluster_by_distance(linkage_table.values, cutoff)
		elif self.cluster_method == 'monocrit':
			clusters = self._cluster_by_monocrit(linkage_table.values, cutoff, inconsistent)

		elif self.cluster_method == 'inconsistent':
			clusters = self._cluster_by_inconsistent(linkage_table.values, cutoff, inconsistent)
		else:
			message = f"'{self.cluster_method}' is not a currently supported clustering method when using hierarchical clustering."
			logger.error(message)
			raise ValueError(message)

		if labels is not None:
			clusters = self._label_clusters(clusters, labels)

		return clusters
	@staticmethod
	def adjust_similarity_cutoff(quantile:float, distances: List[float])->float:
		""" Adjusts the `similarity_cutoff` value to work with the distance observations."""

		#result = pandas.Series(distances).quantile(quantile)
		result = pandas.Series([i for i in distances if 0 < i < max(distances)]).quantile(quantile)
		return result


	def run(self, pair_array: DistanceCache, starting_genotypes: List[List[str]] = None, similarity_cutoff: Optional[float] = None) -> projectdata.DataHierarchalCluster:
		"""
		Parameters
		----------
		pair_array
		starting_genotypes: List[List[str]]
			Each element should be a list of trajectories known to be in the same genotype.
		similarity_cutoff: Optional[float]
			If not given, the similarity cutoff will be generated automatically.
		"""

		# If known genotypes are given, modify the pair_array so that they will be grouped together.
		if starting_genotypes:
			pair_array = self._add_starting_genotypes(pair_array, starting_genotypes)
		squaremap = pair_array.squareform()
		distance_array = distance.squareform(squaremap.values)
		linkage_table = self.link_clusters(distance_array, len(squaremap.index))
		reduced_linkage_table = linkage_table[['left', 'right', 'distance', 'observations']]  # Removes the extra column

		if similarity_cutoff is None:
			quantile = 0.05
		else:
			quantile = similarity_cutoff

		distance_cutoff = self.adjust_similarity_cutoff(quantile, list(pair_array.pairwise_values.values()))

		logger.debug(f"Using Hierarchical Clustering with similarity cutoff {distance_cutoff}")

		clusters = self.cluster(reduced_linkage_table, distance_cutoff, squaremap.index)

		result = projectdata.DataHierarchalCluster(
			clusters = clusters,
			table_linkage = linkage_table,
			distance_cutoff = distance_cutoff,
			distance_quantile = quantile
		)

		return result




import matplotlib.pyplot as plt

plt.clf()
# plt.switch_backend('agg')
import pandas
from pathlib import Path
from typing import Dict, Optional, List, Tuple, Union
from muller import widgets
from muller.graphics.palettes import palette_distinctive, Palette
from loguru import logger


class TimeseriesPlot:
	def __init__(self, render: bool = True, legend: bool = True, scale: int = 1, style: Optional[str] = None):
		# Set the default aspect ratio
		self.length_x = 12
		self.length_y = 10
		self.scale = scale
		if self.scale < 1:
			self.scale = 1

		self.dpi = 250

		# Parameters concerning the overall plot
		self.default_color = "#333333"
		self.style = style
		if self.style == 'nature':
			self._set_style_nature()
		else:
			self._set_style_default()

		# Parameters concerning the plot legend.
		self.legend = legend
		self.legend_font_properties = {
			'size': 12 * self.scale
		}  # The size of the font used to label each series in the legend.
		self.legend_location = 'right'
		self.legend_title = 'Genotypes'

		# Set up the fontsizes for each labeltype
		self.label_size_axis, self.label_size_title, self.label_size_ticks = self.set_scale(scale)

	@staticmethod
	def set_scale(scale: int = 1) -> Tuple[int, int, int]:
		# Made staticmethod so that pycharm doesn't complain about object properties being defined outside of __init__()
		label_size_axis = 24 * scale
		label_size_title = 42 * scale
		label_size_ticks = 18 * scale
		return label_size_axis, label_size_title, label_size_ticks

	def _set_style_default(self) -> None:
		# Parameters concerning the overall plot
		self.xaxis_label = "Generation"
		self.yaxis_label = 'Frequency'
		self.background_color = 'white'

		self.markersize = 4 * self.scale  # The size of individual markers in the plot
		self.markertype = 'o'  # The type of marker to use.

		self.linestyle = 'solid'
		self.linewidth = 2 * self.scale

	def _set_style_nature(self):
		""" Configures `TimeseriesPlot` to use a style similar to that in the yeast nature paper. """
		self.xaxis_label = "Generation"
		self.yaxis_label = "Frequency"
		self.background_color = "white"

		self.markertype = 'o'
		self.markersize = 12 * self.scale

		self.linestyle = 'solid'
		self.linewidth = 3 * self.scale

	def _apply_style(self, axes: plt.Axes, plot_title, xmax: Optional[int] = None) -> plt.Axes:
		# Parameters concerning subfeatures of the plot.
		axes.set_ylabel(self.yaxis_label, fontsize = self.label_size_axis)
		axes.set_xlabel(self.xaxis_label, fontsize = self.label_size_axis)
		axes.set_facecolor(self.background_color)
		axes.set_title(plot_title, fontsize = self.label_size_title)
		axes.set_ylim(0,
			1.01)  # The maximum possible value for our data is 100% AKA 1. Leave a little room so the lines at 100% aren't obscured.
		if xmax:
			axes.set_xlim(0, xmax + 1)  # SMall offset so the final point isn't sut off

		if self.style == 'nature':
			axes.set_yticks([0, 0.5, 1])

		axes.tick_params(axis = 'both', labelsize = self.label_size_ticks)

		return axes

	def _initialize_plot(self, ax: Optional[plt.axes]) -> plt.Axes:
		if ax is None:
			fig, ax = plt.subplots(figsize = (self.length_x * self.scale, self.length_y * self.scale))
		return ax

	@staticmethod
	def get_palette(table: pandas.DataFrame) -> Dict[str, str]:
		return palette_distinctive.generate_distinctive_palette(table.index)

	def plot_multiple(self, timeseries: pandas.DataFrame, palettes: List[Palette], ax: Optional[plt.Axes] = None,
			filenames: Optional[List[Path]] = None):
		for palette in palettes:
			fnames = filenames[palette.name]
			self.plot(timeseries, palette, ax, fnames)

	def plot(self, timeseries: pandas.DataFrame, palette: Union[Dict[str, str], Palette] = None,
			ax: Optional[plt.Axes] = None,
			filename: Optional[Path] = None) -> plt.Axes:
		""" Plots a generic timeseries dataframe. The plot labels are inferred from how the index labels are formatted.
			Parameters
			----------
			timeseries: pandas.DataFrame
				A dataframe where each column is a timepoint and each row is a specific series to plot.
			palette: Dict[str,str]
				Maps each series id to the proper color to use.
			ax: Optional[plt.Axes]
				Specifies the plt.Axes object to use.
			filename: Optional[Path]
				The resulting figure will be saved to this filename if it is provided.
		"""
		# Set up the plotting area.
		if palette is None: palette = {}
		self.set_scale()

		ax = self._initialize_plot(ax)
		try:
			plot_title = 'Genotypes' if 'genotype' in timeseries.index[0] else 'Trajectories'
		except TypeError:
			message = f"Could not iterate over the first element of the index. Is the table indexed by genotype?"
			logger.debug(message)
			plot_title = "Timeseries"
		numeric_columns = list(widgets.get_numeric_columns(timeseries.columns))
		timeseries = timeseries[numeric_columns]

		for series_id, series in timeseries.iterrows():
			color = palette.get(series_id, self.default_color)
			# Make sure that the timeseries is in the right order
			# Some datasets may be out of order.
			trajectory_timeseries = sorted((column, series[column]) for column in numeric_columns)
			x_values, y_values = zip(*trajectory_timeseries)

			ax.plot(
				x_values, y_values,
				self.markertype,
				color = color,
				label = series_id,
				marker = self.markertype,
				markersize = self.markersize,
				linewidth = self.linewidth,
				linestyle = self.linestyle
			)
		ax = self._apply_style(ax, plot_title, max(int(i) for i in timeseries.columns))
		ax.set_xlim(0, max(timeseries.columns))
		if self.legend and False:
			legend = ax.legend(
				loc = self.legend_location,
				prop = self.legend_font_properties,
				title = self.legend_title
			)
			legend.get_title().set_fontsize(str(self.legend_font_properties['size']))

		if filename:
			self.save_figure(filename)
		return ax

	def save_figure(self, filename: Path):
		""" Saves the diagram in every format available in self.filetypes"""
		plt.savefig(filename, dpi = self.dpi)


if __name__ == "__main__":
	pass



import math
import statistics
from typing import Dict, List, Tuple

import pandas
import scipy.stats as stats
from loguru import logger

try:
	from muller import widgets
	from muller.inheritance import areascore
	from muller.inheritance import polygon
except ModuleNotFoundError:
	from . import areascore
	from . import polygon


class LegacyScore:
	def __init__(self, pvalue: float, dlimit: float, flimit: float):
		self.pvalue = pvalue
		self.dlimit = dlimit
		self.flimit = flimit

	def calculate_score_greater(self, nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> float:
		"""
			Tests whether the nested genotype is consistenty larger than  the nested genotype. The scoring is as follows:
			`nested_genotype` always larger than `unnested_genotype`: 1
			`nested_genotype` consistently larger than `unnested_genotype`: 0.5
			`nested_genotype` not consistently larger than `unnested_genotype`: 0
			`nested_genotype` sometimes smaller than `unnested_genotype`: -0.5
		Parameters
		----------
		nested_genotype, unnested_genotype: pandas.Series
		"""
		difference = nested_genotype - unnested_genotype

		score = int(difference.sum() > (self.pvalue * len(difference) / 2))

		# There may be a point where the unnested genotype is greater than the nested genotype.

		if any(difference < 0):
			score -= 1
		# logger.debug(f"{unnested_genotype.name}\t{nested_genotype.name}\t{score}")
		return score

	def calculate_derivative_score(self, left: pandas.Series, right: pandas.Series) -> float:
		"""
			Tests whther the two series are correlated or anticorrelated with each other. The scoring is as follows:
			correlated: 2
			uncorrelated: 0
			anticorrelated: -2
		Parameters
		----------
		left, right: pandas.Series
			The two series to test.
		"""
		# Pandas implementation of the derivative check, since it basically just checks for covariance.
		valid_left, valid_right = widgets.get_valid_points(left, right, self.dlimit, self.flimit)
		if valid_left.empty:
			covariance = math.nan
		else:
			covariance = valid_left.cov(valid_right)

		if covariance > 0.01: score = 2
		elif covariance < -0.01: score = -2
		else: score = 0
		return score

	@staticmethod
	def calculate_area(nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> float:
		"""
			Calculates a score based on the probability the unnested genotype is a subset of the nested_genotype.
			This take into account both the area of the nested genotype and the area of all other genotypes not in the nested genotype.
		Parameters
		----------
		nested_genotype, unnested_genotype: pandas.Series
		"""

		# If the nested genotype is not fixed, group the remaining frequencies into an `other` category.
		# noinspection PyTypeChecker
		other_genotypes: pandas.Series = 1 - nested_genotype
		area_nested = areascore.area_of_series(nested_genotype)
		area_unnested = areascore.area_of_series(unnested_genotype)
		area_other = areascore.area_of_series(other_genotypes)
		common_area_nested = areascore.calculate_common_area(nested_genotype, unnested_genotype)
		common_area_other = areascore.calculate_common_area(other_genotypes, unnested_genotype)

		is_subset_nested = areascore.is_subset_legacy(area_nested, area_unnested, common_area_nested)
		is_subset_other = areascore.is_subset_legacy(area_other, area_unnested, common_area_other)

		if is_subset_nested and is_subset_other:
			score = int(common_area_nested > 2 * common_area_other)
		elif is_subset_nested:
			score = 2
		else:
			score = -2
		# logger.debug(f"{unnested_genotype.name}\t{nested_genotype.name}\t{score}")
		return score

	def calculate_summation_score(self, left: pandas.Series, right: pandas.Series) -> int:
		"""
			Tests whether two genotypes consistently sum to a value greater than the fixed breakpoint. This suggests that one of the genotypes
			is in the background of the other, since otherwise the maximum combined frequency should, at most, be equal to the fixed cutoff value.
		Parameters
		----------
		left, right: pandas.Series
		"""
		combined = left + right
		above_fixed = combined - self.flimit

		# Finally, check whether the frequency of this series is consistently above the `cutoff` value.
		# Timepoints that do not sum to greater than the fixed breakpoint will be negative.
		result = above_fixed.mean() > self.pvalue

		return int(result)


class Score:
	""" Refactored as a class so that all the dlimit,flimit,pvalue,etc variables don't have to be passed around"""

	def __init__(self, dlimit: float, flimit: float, pvalue: float, weights:Tuple[int,int,int,int] = (1, 2, 1, 2)):
		self.pvalue = pvalue
		self.dlimit = dlimit
		self.flimit = flimit
		self.slimit = 0.15

		# Keep this around as a fallback
		# Shapely has been having issues, so may need to fallback to the legacy area score.
		self.legacy_scorer = LegacyScore(pvalue, dlimit, flimit)
		self.weight_greater = weights[0]
		self.weight_above_fixed = weights[1]
		self.weight_derivative = weights[2]
		self.weight_jaccard = weights[3]

		self.debug = False

	def _get_growth_regions(self, series: pandas.Series) -> List[bool]:
		""" Returns True for regions of positive growth and fixed."""
		difference_series = series.diff()

		boolseries = [(i > self.flimit or j > 0) for i, j in zip(series.values, difference_series.values)]
		return boolseries

	def calculate_score_greater(self, nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> float:
		use_advanced = False
		series_overlap = widgets.overlap(nested_genotype, unnested_genotype, self.dlimit)
		if series_overlap == 0:
			return self.weight_greater * -1

		# THe t-test has to be corrected for the case where the two series do not completely overlap
		nested_genotype, unnested_genotype = widgets.get_valid_points(nested_genotype, unnested_genotype, self.dlimit)
		if use_advanced:
			raise NotImplementedError
		else:
			score = self.calculate_score_greater_basic(nested_genotype, unnested_genotype)

		return float(score)  # Cast to float so the dtypes are consistent

	def calculate_score_greater_basic(self, nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> int:
		difference = nested_genotype - unnested_genotype

		nested_is_above_unnested = difference > self.dlimit
		unnested_is_above_nested = difference < -self.dlimit

		# Select the timepoints that correspond to whichever was greater.
		nested = difference[nested_is_above_unnested]
		unnested = difference[unnested_is_above_nested]

		portion_of_series = 0.5
		# The `cutoff` value is just half the total number of timepoints.
		cutoff = len(difference) * portion_of_series

		nested_timepoints = len(nested) # Number of timepoints where the nested genotype was greater
		unnested_timepoints = len(unnested) # Number of timepoints where the unnested genotype was greater.
		# This is adding up the difference in values. Basically the area under the curve.
		# Since each series already omits the timepoints not in that category the abs() method just
		# makes sure the differences are positive.
		nested_sum = abs(nested.sum())
		unnested_sum = abs(unnested.sum())

		# This is basically the mean of the difference between series.
		nested_mean = abs(nested.mean())
		unnested_mean = abs(unnested.mean())

		# Make sure that even if nested is generally larger than unnested, that unnested was not also greater than nested significantly.

		is_consistently_greater = (nested_timepoints > cutoff) and (nested_mean > self.dlimit)
		is_significantly_greater = nested_sum > self.slimit
		left = is_significantly_greater or is_consistently_greater

		is_consistently_less = (unnested_timepoints > cutoff) and (unnested_mean > self.dlimit)
		is_significantly_less = unnested_sum > self.slimit
		right = is_significantly_less or is_consistently_less

		if left and right:
			score = math.nan
		elif is_consistently_greater or is_significantly_greater:
			score = 1
		elif is_consistently_less or is_significantly_less:
			score = -1
		else:
			score = math.nan

		return score

	def _single_sample_ttest(self, left:pandas.Series, right:pandas.Series):
		""" Implements a single sample ttest."""
		combined_series = (left + right) - (1+self.dlimit)
		logger.debug(combined_series.tolist())
		# Test if the result is greater than 0
		statistic, pvalue = stats.ttest_1samp(combined_series.values, 0)
		return statistic, pvalue
	def _multiple_sample_ttest(self, left:pandas.Series, right:pandas.Series):
		combined_series = (left + right).tolist()[1:]

		mean_c = statistics.mean(combined_series)
		var_c = self.dlimit
		mean_f = 1 + self.dlimit
		var_f = self.dlimit

		forward_statistic, forward_pvalue = stats.ttest_ind_from_stats(
			mean1 = mean_c,
			std1 = var_c ** 2,
			nobs1 = len(combined_series),
			mean2 = mean_f,
			std2 = var_f ** 2,
			nobs2 = len(combined_series)
		)
		return forward_statistic, forward_pvalue

	def calculate_score_above_fixed(self, left: pandas.Series, right: pandas.Series) -> int:
		"""
			Tests whether two genotypes consistently sum to a value greater than the fixed breakpoint. This suggests that one of the genotypes
			is in the background of the other, since otherwise the maximum combined frequency should, at most, be equal to the fixed cutoff value.
			Keep in mind that the variance is defined as the uncertainty in the measurements rather than computed using the given values.
		Parameters
		----------
		left, right: pandas.Series
		"""
		# Including points where one genotype was not detected will skew the results.
		left, right = widgets.get_valid_points(left, right, dlimit = self.dlimit, inner = True)
		combined_series = (left + right).tolist()[1:]

		if len(combined_series) == 0:
			result = 0
		elif len(combined_series) == 1:
			result = combined_series[0] > self.flimit
		else:
			forward_statistic, forward_pvalue = self._multiple_sample_ttest(left, right)
			#forward_statistic, forward_pvalue = self._single_sample_ttest(left, right)
			# Since we're using a two-sided test we need to convert it to a one-sided test.
			result = forward_pvalue / 2 < self.pvalue and forward_statistic > 0

		return int(result)

	def calculate_score_area(self, nested_genotype: pandas.Series, unnested_genotype: pandas.Series) -> float:
		"""
			Calculates a score based on the probability the unnested genotype is a subset of the nested_genotype.
			This take into account both the area of the nested genotype and the area of all other genotypes not in the nested genotype.
		Parameters
		----------
		nested_genotype, unnested_genotype: pandas.Series

		"""

		# If the nested genotype is not fixed, group the remaining frequencies into an `other` category.
		difference_series = nested_genotype - unnested_genotype
		if difference_series.mean() > 0:
			# noinspection PyTypeChecker
			other_genotypes: pandas.Series = self.flimit - nested_genotype
		else:
			# noinspection PyTypeChecker
			other_genotypes: pandas.Series = self.flimit - unnested_genotype  # In case we're testing if a small genotype contains a large genotype

		other_genotypes = other_genotypes.mask(lambda s: s < 0, 0.0001)  # Since the flimit is not exactly 1.

		unnested_polygon = polygon.as_polygon(unnested_genotype)

		nested_polygon = polygon.as_polygon(nested_genotype)
		other_polygon = polygon.as_polygon(other_genotypes)

		is_subset_nested = areascore.is_subset_polygon(nested_polygon, unnested_polygon)
		is_subset_other = areascore.is_subset_polygon(other_polygon, unnested_polygon)
		is_subset_nested_reversed = areascore.is_subset_polygon(unnested_polygon, nested_polygon)  # Check the reverse case

		nested_area = areascore.area_of_series(nested_genotype)
		unnested_area = areascore.area_of_series(unnested_genotype)
		common_area_nested = areascore.X_and_Y_polygon(unnested_polygon, nested_polygon)
		xor_area_unnested = areascore.difference_polygon(unnested_polygon, nested_polygon)  # This does not distinguish between xor left vs xor right


		if self.debug:
			logger.debug(
				f"calculate_score_jaccard()->({is_subset_nested}, {is_subset_other}, {is_subset_nested_reversed}), ({common_area_nested:.2f}, {xor_area_unnested:.2f})")
		if is_subset_nested and is_subset_other:
			# Evidence for both scenarios
			# Test if the nested genotype is sufficiently large to assume the unnested genotype is a subset.
			# Test only the area where the unnested genotype was detected.
			other_polygon = polygon.as_polygon(other_genotypes)
			common_area_other = areascore.X_and_Y_polygon(unnested_polygon, other_polygon)
			score = int(common_area_nested > 2 * common_area_other)

		elif is_subset_nested:
			score = 1
		elif is_subset_nested_reversed and not is_subset_other:
			score = -1
		else:
			score = 0
		if score == 0 and xor_area_unnested > common_area_nested * 2:
			score = -1
		elif unnested_area > 2 * nested_area:
			score = -1
		score = score * self.weight_jaccard
		return score

	def calculate_score_derivative(self, left: pandas.Series, right: pandas.Series) -> float:
		"""
			Tests whther the two series are correlated or anticorrelated with each other. The scoring is as follows:
			correlated: 2
			uncorrelated: 0
			anticorrelated: -2
		Parameters
		----------
		left, right: pandas.Series
			The two series to test.
		"""
		# Pandas implementation of the derivative check, since it basically just checks for covariance.
		valid_left, valid_right = widgets.get_valid_points(left, right, self.dlimit, self.flimit, inner = True)

		if valid_left.empty:
			score = 0
		elif len(valid_left) > 20 or True:
			dotproduct, correlated_timepoints = self.derivative(valid_left, valid_right)
			#logger.debug(f"{dotproduct}, {correlated_timepoints}, {len(valid_left)}")
			#logger.debug(valid_left.diff().tolist())
			#logger.debug(valid_right.diff().tolist())
			if dotproduct > 0.01:
				score = 1
			elif dotproduct < -0.01:
				score = -1
			else:
				score = 0
		else:
			covariance = valid_left.cov(valid_right)
			if covariance > 0.01: score = 1
			elif covariance < -0.01: score = -1
			else: score = 0
		score = score * self.weight_derivative
		return score

	def score_pair(self, nested_genotype: pandas.Series, unnested_trajectory) -> Dict[str, float]:
		detected_left, detected_right = widgets.get_valid_points(nested_genotype, unnested_trajectory, dlimit = self.dlimit, inner = False)
		if self.debug:
			logger.debug(f"Scoring a pair of series:")
			logger.debug(f"\t{nested_genotype.name}\t{nested_genotype.values}\t{nested_genotype.index}")
			logger.debug(f"\t{unnested_trajectory.name}\t{unnested_trajectory.values}\t{unnested_trajectory.index}")
			logger.debug(f"{self.dlimit}, {self.flimit}, {self.pvalue}")
			logger.debug(f"The detected portion of the series: ")
			logger.debug(f"\t{detected_left.values}")
			logger.debug(f"\t{detected_right.values}")

		if len(detected_left) < 3:
			score_fixed = self.legacy_scorer.calculate_summation_score(detected_left, detected_right)
		else:
			score_fixed = self.calculate_score_above_fixed(detected_left, detected_right)

		score_greater = self.calculate_score_greater(detected_left, detected_right)
		if math.isnan(score_greater): score_greater = 0
		score_area = self.calculate_score_area(nested_genotype, unnested_trajectory)

		total_score = score_fixed + score_greater + score_area
		if self.debug:
			logger.debug(f"{nested_genotype.name}\t{unnested_trajectory.name}\t{score_fixed}\t{score_greater}\t{score_area}\t{total_score}")
		if total_score > 0:
			# The derivative check is only useful when deciding between possible candidates, since it does not provide evidence itself that a
			# genotype is a potential background. So, at least one of the other checks should have been passed with no
			# evidence against the candidate background.

			# The derivative score should only be computed using the timepoints where the series overlap.
			detected_left, detected_right = widgets.get_valid_points(nested_genotype, unnested_trajectory, dlimit = self.dlimit, inner = True)
			score_derivative = self.calculate_score_derivative(detected_left, detected_right)
			# Note that a previous version accidentlly added the derivative cutoff to the total score.
			total_score += score_derivative
		else:
			score_derivative = math.nan
		if math.isnan(score_derivative): score_derivative = 0
		score_data = {
			'nestedGenotype':   nested_genotype.name,
			'unnestedGenotype': unnested_trajectory.name,
			'scoreGreater':     score_greater,
			'scoreFixed':       score_fixed,
			'scoreArea':        score_area,
			'scoreDerivative':  score_derivative,
			'totalScore':       total_score
		}
		return score_data

	def derivative(self, left: pandas.Series, right: pandas.Series) -> Tuple[float, int]:
		# l, r = widgets.get_valid_points(left, right, 0.03, 0.97, inner = True)
		l, r = left, right
		normalize = lambda s: 1 if s > self.dlimit else (-1 if s < -self.dlimit else 0)

		ldiff = l.diff()[1:]
		rdiff = r.diff()[1:]

		result = ldiff.dot(rdiff)

		ldiff = ldiff.apply(normalize)
		rdiff = rdiff.apply(normalize)

		nresult = ldiff.dot(rdiff)
		return result, nresult



#!python
"""
	Main script to run the muller workflow.
	.
	|---- .clade.annotated.(png|svg)
	|---- .genotypes.tsv
	|---- .lineage.clade.(png|svg)
	|---- figures/
	|----|---- clade/
	|----|----|---- .clade.annotated.(png|svg)
	|----|----|---- .clade.unannotated.(png|svg)
	|----|----|---- .clade.genotypes.(png|svg)
	|----|----|---- .clade.trajectories.(png|svg)
	|----|---- unique/
	|----|----|---- .unique.annotated.(png|svg)
	|----|----|---- .unique.unannotated.(png|svg)
	|----|----|---- .unique.genotypes.(png|svg)
	|----|----|---- .unique.trajectories.(png|svg)
	|----|---- .dendrogram.png
	|----|---- .pairwisedistance.svg
	|---- scripts/
	|----|---- .r
	|---- suplementary-files/
	|----|---- .genotypeinformation.json
	|----|---- .lineagescores.tsv
	|----|---- .options.json
	|---- tables/
	|----|---- .trajectories.tsv
	|----|---- .genotypes.tsv
	|----|---- .distance.tsv
	|----|---- .linkagematrix
	|----|---- .ggmuller.edges.tsv
	|----|---- .ggmuller.populations.tsv
	|----|---- .mullerdataframe.tsv
"""
import argparse
from pathlib import Path
from typing import *

import pandas

from muller import graphics, widgets
from muller.graphics import graphicsio

pandas.set_option('mode.chained_assignment',
	None)  # This disables the warning about setting a value on a copy of a dataframe.
from loguru import logger

from muller import clustering, dataio, inheritance, commandline_parser
from muller.dataio import projectdata, annotations, projectpaths

logger.remove()  # Need to remove the default sink so that the logger doesn't print messages twice.
import sys

if commandline_parser.DEBUG:
	logger.add(sys.stderr, level = "DEBUG")
else:
	logger.add(sys.stderr, level = 'INFO', format = "{time:YYYY-MM-DD HH:mm:ss} {level} {message}")


def run_genotype_inference_workflow(trajectoryio: Union[str, Path, pandas.DataFrame], metric: str, dlimit: float,
		flimit: float,
		similarity_cutoff: float, known_genotypes: Optional[Path] = None, threads: Optional[int] = None,
		is_genotype: bool = False) -> projectdata.DataGenotypeInference:
	"""
	Parameters
	----------
	trajectoryio: Union[str,Path,pandas.DataFrame]
		Either the path to a table with mutational trajectories or the table itself.
		The table should already be cleaned/filtered.
		The table should be organized with timepoints as columns and rows as genotype frequencies.
		Ex.
		genotype	0	7	13 ...
		genotype-1	0	.1	.3 ...
		genotype-2	0.2	.1 ...
	metric: Literal['']
	dlimit, flimit: float
	similarity_cutoff: float
	known_genotypes
	threads
	is_genotype: bool
	"""
	if isinstance(trajectoryio, (str, Path)):
		logger.info(f"Reading '{trajectoryio}' as the trajectory table.")
		trajectories, trajectory_info = dataio.parse_trajectory_table(trajectoryio)
	else:
		trajectories = trajectoryio
		trajectory_info = None

	# Read in the starting genotypes.
	if known_genotypes:
		known_genotypes = dataio.parse_known_genotypes(known_genotypes)

	genotype_generator = clustering.ClusterMutations(
		metric = metric,
		dlimit = dlimit,
		flimit = flimit,
		starting_genotypes = known_genotypes,
		threads = threads
	)
	if is_genotype:
		logger.info(f"Skipping genotype inference...")
		table_genotypes = trajectories.copy(deep = True)
		table_genotypes = genotype_generator.organizer.run(table_genotypes)
		# Make sure the "Genotype" column is properly named.
		genotype_members = {f"genotype-{k}": [k] for k in table_genotypes.index}
		# Make sure the genotype names are prefixed with 'genotype'
		table_genotypes.index = [f'genotype-{i}' for i in table_genotypes.index]
		table_genotypes.index.name = "Genotype"
		genotype_data = projectdata.DataGenotypeInference(
			table_trajectories = trajectories,
			table_genotypes = table_genotypes,
			genotype_members = genotype_members,
			matrix_distance = None,
			clusterdata = None,
			table_trajectories_info = None
		)
	else:
		genotype_data = genotype_generator.run(trajectories, distance_cutoff = similarity_cutoff)
	genotype_data.table_trajectories_info = trajectory_info
	return genotype_data


def run_genotype_lineage_workflow(genotypeio: Union[str, Path, pandas.DataFrame], dlimit: float, flimit: float,
		pvalue: float, known_ancestry: Optional[Path], conservative: bool) -> projectdata.DataGenotypeLineage:
	"""

	Parameters
	----------
	genotypeio
	dlimit
	flimit
	pvalue
	known_ancestry
	conservative

	Returns
	-------

	"""
	lineage_generator = inheritance.LineageWorkflow(
		dlimit = dlimit,
		flimit = flimit,
		pvalue = pvalue,
		conservative = conservative
	)

	# Read in the input data if it is not already a pandas.DataFrame object
	if isinstance(genotypeio, (str, Path)):
		logger.info(f"Reading '{genotypeio}' as the genotype table.")
		genotypes = dataio.parse_genotype_table(genotypeio)
	else:
		genotypes = genotypeio

	if known_ancestry:
		known_ancestry = dataio.read_map(known_ancestry)
		# Remove the table headers.
		if 'Parent' in known_ancestry:
			known_ancestry.pop('Parent')
		if 'Identity' in known_ancestry:
			known_ancestry.pop('Identity')

	lineage_data = lineage_generator.run(genotypes, known_ancestry)
	return lineage_data


def run_lineageplot_workflow(edgesio: Union[str, Path, pandas.DataFrame], filename: Path,
		sheet_name: Optional[str] = None):
	""" Generates a lineage plot given an `edges` table. The columns should be named `parent` and `identity`.
		An optional `annotation` column will be used to annotate the plot.
	"""
	from muller.graphics import flowchart
	from muller.graphics.palettes import generate_palette
	if isinstance(edgesio, (str, Path)):
		edges = dataio.import_table(edgesio, sheet_name = sheet_name)
	else:
		edges = edgesio
	# Convert the column labels to lowercase
	edges.columns = [i.capitalize() for i in edges.columns]
	edges = edges.set_index('Identity')
	if 'Annotation' in edges.columns:
		edges_annotations = edges.pop("Annotation").to_dict()
		# Make sure 'annotations' is a list of str which is expected by the flowchart
		edges_annotations = {k: [v] for k, v in edges_annotations.items()}
	else:
		edges_annotations = dict()

	palette = generate_palette(edges['Parent'], kind = 'lineage')

	flowchart(edges, palette, edges_annotations, filename)


def get_base_filename(filename: Path, name: Optional[str], sheetname: Optional[str]) -> str:
	if name:
		base_filename = name
	else:
		base_filename = filename.stem
	if sheetname and sheetname != 'Sheet1':
		base_filename += '.' + sheetname

	return base_filename


def run_workflow(program_options: argparse.Namespace):
	# TODO: Test whether the lineage makes sense by computing the sum of genotypes/lineages at each timepoint,
	# Where 100% should be the maximum value.

	"""
		TODO: Running this command will fail:
		python /home/cld100/Documents/github/muller_diagrams/lolipop lineage
		--input traverse-etal-B1-mutationfrequencies.xlsx
		--sheetname trajectories.filtered
		--output traverse-genotypes
		--known-ancestry known_ancestry2.txt
		--genotypes
	"""

	logger.info("Running with the following parameters")
	for key, value in vars(program_options).items():
		logger.info(f"\t{key:<30}{value}")
	output_folder = program_options.output_folder
	paths = projectpaths.OutputFilenames(output_folder, program_options.filename.stem)
	logger.info("Parsing options...")
	data_basic = projectdata.DataWorkflowBasic(
		version = commandline_parser.__VERSION__,
		filename = program_options.filename,
		program_options = program_options
	)
	if not data_basic.program_options.output_folder.exists():
		data_basic.program_options.output_folder.mkdir()
	logger.info("Importing trajectories...")
	trajectory_table, trajectory_info = dataio.parse_trajectory_table(program_options.filename,
		program_options.sheetname)

	# Need to read in the input dataset.
	logger.info("Generating genotypes...")
	result_genotype_inference = run_genotype_inference_workflow(
		trajectory_table,
		program_options.metric,
		dlimit = program_options.dlimit,
		flimit = program_options.flimit,
		similarity_cutoff = program_options.similarity_cutoff,
		known_genotypes = program_options.known_genotypes,
		threads = program_options.threads,
		is_genotype = program_options.is_genotype
	)

	if result_genotype_inference.table_trajectories_info is None:
		result_genotype_inference.table_trajectories_info = trajectory_info

	genotype_annotations = annotations.read_genotype_annotations(trajectory_info,
		result_genotype_inference.genotype_members)

	result_genotype_lineage = run_genotype_lineage_workflow(
		result_genotype_inference.table_genotypes,  # should already be sorted.
		dlimit = program_options.dlimit,
		flimit = program_options.flimit,
		pvalue = program_options.pvalue,
		known_ancestry = program_options.known_ancestry,
		conservative = program_options.conservative
	)

	# Generate the tables needed for generating the muller plots.
	result_ggmuller = run_workflow_ggmuller(
		table_genotypes = result_genotype_inference.table_genotypes,
		series_edges = result_genotype_lineage.clusters.as_ancestry_table(),
		dlimit = data_basic.program_options.dlimit,
		smooth_values = data_basic.program_options.smooth_plot
	)

	paths.save_projectdata_basic(data_basic)
	paths.save_workflow_clustering(result_genotype_inference)
	paths.save_workflow_hierarchy(result_genotype_inference.clusterdata)
	paths.save_workflow_lineage(result_genotype_lineage)
	paths.save_workflow_ggmuller(result_ggmuller)
	# save_tables(data_basic, result_genotype_inference, result_genotype_lineage, genotype_annotations)
	# Save using the older graphics workflow for now.
	render_graphics(
		paths = paths,
		data_basic = data_basic,
		data_inference = result_genotype_inference,
		data_ggmuller = result_ggmuller,
		genotype_annotations = genotype_annotations
	)

	data_basic.save(output_folder)


def run_workflow_ggmuller(table_genotypes: pandas.DataFrame, series_edges: pandas.Series, dlimit: float,
		smooth_values: bool):
	""" Generates the population and edges table used in the muller plot.
		Parameters
		----------
		table_genotypes: pandas.DataFrame
		series_edges: pandas.Series
			The index corresponds the the id of the child genotype, while the linked value is the id of the
			predicted parent genotype.
		dlimit:float
		smooth_values:bool
	"""

	generator_table_population = dataio.GGMuller(cutoff_detection = dlimit, adjust_populations = True)
	table_population = generator_table_population.generate_ggmuller_population_table(series_edges, table_genotypes)
	generator_table_muller = dataio.GenerateMullerDataFrame()
	table_muller = generator_table_muller.run(series_edges, table_population.copy(deep = True))

	result = projectdata.DataGGmuller(
		table_populations = table_population,
		series_edges = series_edges,
		table_muller = table_muller
	)

	return result


def save_tables(data_basic: projectdata.DataWorkflowBasic, data_inference: projectdata.DataGenotypeInference,
		data_lineage: projectdata.DataGenotypeLineage):
	"""
		.figures
		|---- tables
		|----|---- genotypes.tsv
		|----|---- trajectories.tsv
		|----|---- trajectores.filtered.tsv
		|----|---- edges.tsv
		|----|---- populations.tsv
		|----|---- distances.tsv
		|----|---- scorerecords.tsv
		|----|---- linkagetable.tsv
		|----|---- parameters.json
		|----|---- genotypeinformation.json
		|---- scripts
		|----|---- lineage.r
	"""
	delimiter = "\t"
	suffix = "tsv"
	prefix = get_base_filename(data_basic.program_options.filename, data_basic.program_options.name,
		data_basic.program_options.sheetname)
	folder_tables = widgets.checkdir(data_basic.program_options.output_folder / "tables")
	folder_data = widgets.checkdir(data_basic.program_options.output_folder / "data")
	data_basic.save(folder_data)
	data_inference.save(folder_tables, prefix)
	data_lineage.save(folder_tables, prefix)
	"""
	filename_table_genotypes = folder_table / f"{prefix}.genotypes.{suffix}"
	filename_table_trajectories = folder_table / f"{prefix}.trajectories.{suffix}"
	filename_table_trajectories_filtered = folder_table / f"{prefix}.trajectories.filtered.{suffix}"
	filename_table_edges = folder_table / f"{prefix}.edges.{suffix}"
	filename_table_population = folder_table / f"{prefix}.population.{suffix}"
	filename_table_distances = folder_table / f"{prefix}.distances.{suffix}"
	filename_table_scorerecords = folder_table / f"{prefix}.scorerecords.{suffix}"
	filename_table_linkage = folder_table / f"{prefix}.linkage.{suffix}"

	filename_data_parameters = folder_data / f"{prefix}.parameters.json"
	filename_data_genotypes = folder_data / f"{prefix}.genotypes.json"

	data_inference.table_genotypes.to_csv(filename_table_genotypes, sep = delimiter)
	data_inference.table_trajectories.to_csv(filename_table_trajectories, sep = delimiter)
	data_inference.matrix_distance.squareform().to_csv(filename_table_distances, sep = delimiter)
	data_inference.clusterdata.table_linkage.to_csv(filename_table_linkage, sep = delimiter)
	
	data_lineage.table_edges.to_csv(filename_table_edges, sep = delimiter)
	data_lineage.table_populations.to_csv(filename_table_population, sep = delimiter)
	data_lineage.table_scores.to_csv(filename_table_scorerecords, sep = delimiter)
	"""


def render_graphics(paths: projectpaths.OutputFilenames, data_basic: projectdata.DataWorkflowBasic,
		data_inference: projectdata.DataGenotypeInference,
		data_ggmuller: projectdata.DataGGmuller, genotype_annotations: Dict[str, List[str]]):
	""" Graphics are parametrized by filename suffix (png vs svg) and palette.
		.figures
		|---- dendrogram.png
		|---- heatmap.png
		|---- distancedistribution.png
		|---- unique palette
		|    |---- muller.annotated.(svg|png)
		|    |---- muller.unannotated.(svg|png)
		|    |---- timeseriespanel.(svg|png)
		|    |---- lineageplot.(svg|png)
		|---- lineage palette
		|    |---- muller.annotated.(svg|png)
		|    |---- muller.unannotated.(svg|png)
		|    |---- timeseriespanel.(svg|png)
		|    |---- lineageplot.(svg|png)
	"""
	custom_palette = dataio.read_map(data_basic.program_options.genotype_palette_filename)
	prefix = get_base_filename(
		data_basic.program_options.filename,
		data_basic.program_options.name,
		data_basic.program_options.sheetname
	)
	workflow_graphics = graphicsio.OutputGeneratorGraphics(
		project_folder = paths.folder_output,
		palettes = [],
		sample_basename = prefix,
		render = True
	)

	# Plot the figures that aren't parametrized
	if data_inference.clusterdata is not None:
		workflow_graphics.generate_dendrogram(data_inference.clusterdata.table_linkage, data_inference.matrix_distance,
			paths.filename_figure_linkage_plot)
	if data_inference.matrix_distance is not None:
		workflow_graphics.generate_heatmap(data_inference.matrix_distance.squareform(),
			paths.filename_figure_distance_heatmap)
	if data_inference.clusterdata is not None and data_inference.matrix_distance is not None:
		graphics.generate_distance_plot(
			data_inference.matrix_distance.values,
			data_inference.clusterdata.distance_cutoff,
			paths.filename_figure_distribution
		)
	# Set up the generators
	generator_panel_timeseries = graphics.TimeseriesPanel(render = data_basic.program_options.render)
	generator_plot_timeseries = graphics.TimeseriesPlot(render = data_basic.program_options.render)
	generator_plot_muller = graphics.MullerPlot(
		outlines = data_basic.program_options.draw_outline,
		render = data_basic.program_options.render

	)
	generator_plot_muller_panel = graphics.MullerPanel()
	# These figures need to be parametrized by palette type and filetype.
	for palette_name in ["unique", "lineage"]:
		folder_palette = paths.folder_figures_lineage if palette_name == 'lineage' else paths.folder_figures_unique

		current_palette_data = graphics.palettes.generate_palette(
			edges = data_ggmuller.series_edges,
			custom_palette = custom_palette,
			annotations = genotype_annotations,
			kind = palette_name,
		)

		current_palette = graphics.Palette(
			name = palette_name,
			palette = current_palette_data,
			members = data_inference.genotype_members
		)

		filename_palette = paths.filename_palette.with_name(f"palette.{palette_name}.json")
		current_palette.save(filename_palette)

		for suffix in ["png", "svg"]:
			filename_mullerplot_annotated = paths.get_template(folder_palette,
				paths.template_figure_muller_diagram_annotated, suffix)
			filename_mullerplot_unannotated = paths.get_template(folder_palette,
				paths.template_figure_muller_diagram_unannotated, suffix)
			filename_timeseries_panel = paths.get_template(folder_palette, paths.template_figure_panel_timeseries,
				suffix)
			filename_timeseries_genotypes = paths.get_template(folder_palette,
				paths.template_figure_timeseries_genotype, suffix)
			# filename_mullerplot_annotated = folder_palette / f"{prefix}.muller.annotated.{suffix}"
			# filename_mullerplot_unannotated = folder_palette / f"{prefix}.muller.unannotated.{suffix}"
			# filename_timeseries_panel = folder_palette / f"{prefix}.timeseriespanel.{suffix}"
			# filename_timeseries_genotypes = folder_palette / f"{prefix}.timeseries.genotypes.{suffix}"
			# filename_lineageplot = folder_palette / f"{prefix}.lineageplot.{suffix}"
			# filename_muller_panel = folder_palette / f"{prefix}.mullerpanel.{suffix}"
			filename_graphviz_script = folder_palette / f"{prefix}.lineagescript.{suffix}.dot"
			filename_ggmuller_script = folder_palette / f"{prefix}.ggmuller.r"
			filename_lineage_panel = folder_palette / f"{prefix}.lineagepanel.{suffix}"

			filename_lineageplot = paths.get_template(folder_palette, paths.template_figure_lineageplot, suffix)
			filename_muller_panel = paths.get_template(folder_palette, paths.template_figure_panel_muller, suffix)
			# filename_script_graphviz = paths.filename_script_graphviz

			logger.info("Generating the panels...")

			generator_plot_timeseries.plot(
				data_inference.table_genotypes,
				palette = current_palette.get_genotype_palette(),
				filename = filename_timeseries_genotypes
			)

			generator_panel_timeseries.run(
				timeseries_genotype = data_inference.table_genotypes,
				timeseries_trajectory = data_inference.table_trajectories,
				filename = filename_timeseries_panel,
				palette = current_palette,
			)
			logger.info("Generating the annotated muller plots...")
			generator_plot_muller.plot(
				muller_df = data_ggmuller.table_muller,
				color_palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_mullerplot_annotated
			)

			generator_plot_muller_panel.plot(
				timeseries = data_inference.table_genotypes,
				muller_df = data_ggmuller.table_muller,
				palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_muller_panel
			)

			logger.info("Generating the unannotated muller plots...")
			generator_plot_muller.plot(
				muller_df = data_ggmuller.table_muller,
				color_palette = current_palette.get_genotype_palette(),
				annotations = None,
				filename = filename_mullerplot_unannotated
			)
			logger.info("Generating the lineageplot...")
			lineageplot = graphics.flowchart(
				edges = data_ggmuller.series_edges.reset_index(),
				palette = current_palette.get_genotype_palette(),
				annotations = genotype_annotations,
				filename = filename_lineageplot,
				add_score = False
			)

			# Save the scripts
			script_contents_ggmuller = dataio.generate_r_script(
				trajectory = paths.filename_table_trajectories,
				population = paths.filename_table_population,
				edges = paths.filename_table_edges,
				color_palette = current_palette.get_genotype_palette(),
				genotype_labels = list(data_inference.table_genotypes.index)
			)
			filename_ggmuller_script.write_text(script_contents_ggmuller)
			filename_graphviz_script.write_text(lineageplot.to_string())
			if suffix == 'png':
				# PIL doesn't work with svgs
				logger.info("Generating the lineage panel...")
				graphics.generate_lineage_panel(filename_mullerplot_annotated, filename_lineageplot,
					filename_lineage_panel)



import collections
import itertools
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from loguru import logger


@dataclass
class Genotype:
	""" Used to consolidate known information for each genotype. Should probably be replaced by a dict later."""
	label: str
	color_clade: str
	color_custom: Optional[str]
	color_unique: str
	annotations: List[str]
	members: List[str]
	parent: Optional[str] = None


class GenotypeCollection(collections.UserDict):
	""" Provides methods for reproducing the old genotype-to-attribute maps"""

	def get_palette(self, which: str):
		return {k: (getattr(v, which) if v.color_custom is None else v.color_custom) for k, v in self.items()}

	def get(self, which: str) -> Dict[str, Union[str, List[str]]]:
		if 'color' in which:
			return self.get_palette(which)
		return {k: getattr(v, which) for k, v in self.items()}

	def trajectory_palette(self, which: str) -> Dict[str, str]:
		palette = dict()
		genotype_palette = self.get(which)
		for genotype_label, genotype_info in self.items():
			palette.update({m: genotype_palette[genotype_label] for m in genotype_info.members})
		return palette

	def get_annotation_map(self) -> Dict[str, List[Genotype]]:
		data = dict()
		for k, v in self.items():
			for annotation in v.annotations:
				if annotation not in data:
					data[annotation] = [k]
				else:
					data[annotation].append(k)

		return data

	def get_genotype_from_annotation(self, value: str) -> List[Genotype]:
		""" Try to find a given genotype based on the annotations that were extracted
			during the intitial data import."""
		annotation_map = self.get_annotation_map()
		if value in annotation_map:
			return annotation_map[value]
		else:
			# Print a list of similar annotations.
			logger.warning(f"Could not find a genotype matching annotation '{value}'")
			all_annotations = [i.annotations for i in self.values()]
			available = sorted(itertools.chain.from_iterable(all_annotations))
			try:
				from rapidfuzz import process
				similar = process.extract(value, available)
				similar = [i[0] for i in similar if i[1] > 80]
			except ModuleNotFoundError:
				similar = [i for i in available if value in i]
			if similar:
				logger.warning(f"\tDid you mean one of these?")
				for i in similar:
					logger.warning(f"\t\t{i}")
			return []

	def get_trajectory_map(self) -> Dict[str, str]:
		t = dict()
		for label, value in self.items():
			t.update({m: label for m in value.members})
		return t



import json
from pathlib import Path
from typing import Union, Dict, List
import pandas


class OutputFilenames:
	""" Used to organize the files generated by the workflow.
	"""

	def __str__(self):
		string = f"OutputFilenames('{self.folder_output}')"
		return string

	def __init__(self, output: Path, name: str, suffix = 'tsv'):
		self.name = name
		self.suffix = suffix

		def check_folder(path: Union[str, Path]) -> Path:
			path = Path(path)
			if not path.exists():
				path.mkdir()
			return path.absolute()

		self.folder_output = check_folder(output)
		self.folder_supplementary = check_folder(self.folder_output / "supplementary-files")
		self.folder_figures = check_folder(self.folder_output / "graphics")
		self.folder_figures_unique = check_folder(self.folder_figures / "unique")
		self.folder_figures_lineage = check_folder(self.folder_figures / "lineage")
		self.folder_tables = check_folder(self.folder_output / "tables")
		self.folder_scripts = check_folder(self.folder_output / "scripts")

		# General Files
		self.filename_trajectory_table: Path = self.folder_output / (name + f'.trajectories.{suffix}')
		self.filename_table_genotypes: Path = self.folder_output / (name + f'.genotypes.{suffix}')
		self.filename_palette: Path = self.folder_supplementary / (name + f'.palette.json')

		# tables
		self.filename_table_trajectories: Path = self.folder_tables / (name + f'.trajectories.{suffix}')
		self.filename_table_trajectories_rejected: Path = self.folder_tables / (
				name + f"trajectories.rejected.{suffix}")
		# self.filename_table_genotypes: Path = self.folder_tables / (name + f'.genotypes.original.{suffix}')

		self.filename_table_population: Path = self.folder_tables / (name + f'.populations.{suffix}')
		self.filename_table_edges: Path = self.folder_tables / (name + f'.edges.{suffix}')
		self.filename_table_muller: Path = self.folder_tables / (name + f".muller.{suffix}")
		self.filename_table_lineage_scores: Path = self.folder_tables / (name + '.lineagescores.tsv')
		self.filename_table_linkage = self.folder_tables / (name + f".linkagematrix.tsv")
		self.filename_table_distance: Path = self.folder_tables / (name + f".distance.{suffix}")

		# graphics
		# The extensions fo reach figure will be generated based on which file formats the
		# corresponding plotter objects support.
		# The figures will also be generated twice, but there's no point in making a separate Path variable
		# For each pallette used.
		# Muller Plots
		# The muller diagram generator will automatically add filetype extensions and an svg render of each file.

		self.template_figure_panel_timeseries: str = (name + '.timeseriespanel')
		self.template_figure_panel_muller: str = (name + '.mullerpanel')
		self.template_figure_muller_diagram_annotated: str = (name + '.mullerdiagram.annotated')
		self.template_figure_muller_diagram_unannotated: str = (name + '.mullerdiagram.unannotated')
		# self.filename_figure_muller_diagram_clade_annotated: Path = self.folder_figures_lineage / (name + '.annotated')
		# self.filename_figure_muller_diagram_clade_unannotated: Path = self.folder_figures_lineage / (name + '.unannotated')
		# self.filename_figure_muller_diagram_distinct_annotated: Path = self.folder_figures_unique / (name + '.annotated')
		# self.filename_figure_muller_diagram_distinct_unannotated: Path = self.folder_figures_unique / (name + '.unannotated')

		## Timeseries plots
		self.template_figure_timeseries_genotype: str = (name + ".genotypes")
		self.template_figure_timeseries_trajectory: str = (name + f".trajectories")
		# self.filename_figure_timeseries_plot_genotype_unique: Path = self.folder_figures_unique / (name + '.genotypes')
		# self.filename_figure_timeseries_plot_genotype_clade: Path = self.folder_figures_lineage / (name + '.genotypes')

		## Lineage plots
		self.template_figure_lineageplot: str = (name + f".lineage")
		# self.filename_figure_lineage_distinct: Path = self.folder_figures_unique / (name + f".lineage.png")
		# self.filename_figure_lineage_clade: Path = self.folder_output / (name + '.lineage.png')

		## Other plots

		self.filename_figure_distance_heatmap: Path = self.folder_figures / (name + f".pairwisedistance.svg")
		self.filename_figure_linkage_plot = self.folder_figures / (name + f".dendrogram.png")
		self.filename_figure_distribution = self.folder_figures / (name + f".distribution.png")
		# scripts
		self.filename_script_r_script: Path = self.folder_scripts / (name + '.r')
		self.filename_script_graphviz: Path = self.folder_scripts / (name + '.dot')

		# supplementary files
		self.filename_data_genotype_members: Path = self.folder_supplementary / (name + '.genotypemembers.json')
		self.filename_parameters: Path = self.folder_supplementary / (name + '.options.json')
		self.filename_clusterdata: Path = self.folder_supplementary / (name + '.clusterdata.json')
		self.genotype_information: Path = self.folder_supplementary / (name + '.genotypeinformation.json')

	@staticmethod
	def _generate_trajectory_table(table_trajectories: pandas.DataFrame, table_info: pandas.DataFrame,
			genotype_members: Dict[str, List[str]]):
		"""
			Merges all of the available data on trajectories into a single table. Both the `trajectories` and trajectory_info` table
			should be indexed by trajectory id.
		"""
		# Convert the mapping from genotype->members into members->genotype
		trajectory_genotypes = dict()
		for genotype_name, genotype_members in genotype_members.items():
			for member_name in genotype_members:
				trajectory_genotypes[member_name] = genotype_name

		table_trajectories = table_trajectories.merge(table_info, left_index = True, right_index = True, how = 'outer')

		# Add a column to the `trajectories` table with the resulting parent genotype.
		table_trajectories['genotype'] = [trajectory_genotypes.get(i, 'rejected') for i in table_trajectories.index]

		return table_trajectories

	@staticmethod
	def get_template(folder: Path, template: str, extension: str):
		"""
			Generates an actual filename based on the template string.
		"""
		if not extension.startswith('.'):
			extension = '.' + extension
		result = folder / (template + extension)
		return result

	def save_projectdata_basic(self, data):
		program_options = {key: str(value) for key, value in vars(data.program_options).items()}
		options = {
			'version':    data.version,
			'filename':   str(data.filename),
			'parameters': program_options
		}

		self.filename_parameters.write_text(json.dumps(options, indent = 4, sort_keys = True))

	def save_workflow_clustering(self, data):

		table_trajectories = self._generate_trajectory_table(
			data.table_trajectories,
			data.table_trajectories_info,
			data.genotype_members
		)

		table_trajectories.to_csv(self.filename_table_trajectories, sep = self.delimiter)
		# TODO: merge table_trajectories_info with the trajectories table.
		# data.table_trajectories_info.to_csv()
		members = {key: '|'.join(values) for key, values in data.genotype_members.items()}

		data.table_genotypes['members'] = [members[i] for i in data.table_genotypes.index]
		data.table_genotypes.to_csv(self.filename_table_genotypes, sep = self.delimiter)
		if data.matrix_distance is not None:
			data.matrix_distance.squareform().to_csv(self.filename_table_distance, sep = self.delimiter)
		if data.clusterdata is not None:
			data.clusterdata.table_linkage.to_csv(self.filename_table_linkage, sep = self.delimiter)

		self.filename_data_genotype_members.write_text(json.dumps(data.genotype_members, indent = 4, sort_keys = True))

	def save_workflow_hierarchy(self, data):
		if data is not None:
			self.filename_clusterdata.write_text(json.dumps(data.to_dict(), indent = 4, sort_keys = True))

	def save_workflow_lineage(self, data):
		data.table_scores.to_csv(self.filename_table_lineage_scores, sep = self.delimiter, index = False)

	def save_workflow_ggmuller(self, data):
		if not self.filename_table_population.exists():
			data.table_populations.to_csv(self.filename_table_population, sep = self.delimiter, index = False)
		if not self.filename_table_edges.exists():
			table_edges = data.series_edges.to_frame().reset_index()
			# GGmuller expected the columns to be ordered as ['Parent', 'Identity']
			table_edges = table_edges[['Parent', 'Identity']]
			table_edges.to_csv(self.filename_table_edges, sep = self.delimiter, index = False)

		data.table_muller.to_csv(self.filename_table_muller, sep = self.delimiter, index = False)

	@property
	def delimiter(self) -> str:
		if self.suffix == 'tsv':
			return '\t'
		else:
			return ','



from pathlib import Path
from typing import Dict, List, Optional

import pandas
import pygraphviz
from loguru import logger

from muller import widgets


def get_node_label_properties(identity: str, genotype_color: str, annotation: List[str]) -> Dict[str, str]:
	luminance = widgets.calculate_luminance(genotype_color)
	if luminance > 0.5:
		font_color = "#333333"
	else:
		font_color = "#FFFFFF"

	label = [identity] + annotation
	label = "\n".join(label)

	label_properties = {
		'label':     label,
		'fontcolor': font_color
	}
	return label_properties


def flowchart(edges: pandas.DataFrame, palette: Dict[str, str], annotations: Dict[str, List[str]] = None, filename: Optional[Path] = None, add_score:bool = True)->pygraphviz.AGraph:
	"""
		Creates a lineage plot showing the ancestry of each genotype.
	Parameters
	----------
	edges: pandas.DataFrame
		Should have three columns: 'parent', 'identity', 'score'
	palette
	annotations
	filename: Optional[Path]
		Where to save the plot.

	Returns
	-------

	"""
	identity_column = "Identity"
	parent_column = "Parent"

	# Check if the edges table is using the identity column an an index.
	if edges.index.name and edges.index.name == identity_column.lower():
		edges = edges.reset_index()
	if annotations is None:
		annotations = {}

	graph = pygraphviz.AGraph(splines = 'ortho', orientation = 90)
	graph.node_attr['shape'] = 'box'
	graph.node_attr['style'] = 'filled,rounded'
	graph.node_attr['fontname'] = 'lato'
	graph.edge_attr['dir'] = 'forward'

	for identity in edges[identity_column].unique():
		genotype_color = palette.get(identity)
		graph.add_node(identity, color = "#333333", fillcolor = genotype_color)
		node = graph.get_node(identity)
		node_label_properties = get_node_label_properties(identity, genotype_color, annotations.get(identity, []))
		node.attr.update(node_label_properties)

	for index, row in edges.iterrows():
		parent = row[parent_column]
		identity = row[identity_column]
		arguments = {
			'tooltip': f"{parent}",
			'labeldistance': 2.0
		}
		if add_score:
			score = row.get('score', 0)
			arguments['headlabel'] = f"{score:.1f}"
		graph.add_edge(parent, identity, **arguments)

	if filename:
		try:
			graph.draw(str(filename), prog = 'dot')
		except Exception as exception:
			logger.error(f"Could not generate lineage plot: {exception}")
	return graph



""" Implements a class which can compute statistical values over a set of clusters. Implemented as its own class to make is usable elsewhere."""
import itertools
import statistics
from typing import Dict, List, Tuple, Union

from loguru import logger

from muller.clustering.metrics import DistanceCache

ClusterType = List[str]


class ClusterSet:
	""" Provides descriptive statistics for a set of clusters."""
	def __init__(self, clusters: List[ClusterType], distances: Union[DistanceCache, Dict[Tuple[str, str], float]]):
		if not isinstance(distances, DistanceCache):
			self.distances = DistanceCache(distances)
		else:
			self.distances = distances

		self.number_of_points = len(self.distances.squareform().index)

		self.clusters = clusters
		self._genotypes = None  # Cache for the `self.genotypes` property.

	def __len__(self) -> int:
		return len(self.clusters)

	def __getitem__(self, item):
		return self.genotypes[item]

	def calculate_index(self) -> float:
		""" Calculates the CH index for the clusters."""

		B = self.calculate_within_cluster_variations()
		W = self.calculate_between_cluster_variation()
		logger.debug(f"{B}, {W}")
		result = (B / len(self) - 1) / (W / (self.number_of_points - len(self)))
		return result

	def calculate_between_cluster_variation(self):
		clusters = self.get_genotypes(self.clusters)
		combinations = itertools.combinations(clusters.keys(), 2)
		cluster_distances = list()
		for left_genotype, right_genotype in combinations:
			# Need to calculate the distance between these two clusters
			left_cluster = clusters[left_genotype]
			right_cluster = clusters[right_genotype]

			distance = self.calculate_distance_between_clusters(left_cluster, right_cluster, 'min')
			#logger.debug(f"{left_genotype}, {right_genotype}, ({len(left_cluster), len(right_cluster)}), {distance}")
			cluster_distances.append(distance)

		mean_cluster_distance = statistics.mean(cluster_distances)
		variances = [(i - mean_cluster_distance) ** 2 for i in cluster_distances]
		return sum(variances) / len(clusters)

	def calculate_distance_between_clusters(self, left_cluster: ClusterType, right_cluster: ClusterType, method = 'min') -> float:
		# Need to pair the left trajectories with the right trajectories, but not left to left.

		between_cluster_distances = [self.distances[left, right] for right in right_cluster for left in left_cluster]

		if method == 'min':
			return min(between_cluster_distances)
		elif method == 'max':
			return max(between_cluster_distances)
		elif method == 'mean':
			return statistics.mean(between_cluster_distances)
		else:
			message = f"Invalid method for calculating the between-cluster distance: '{method}'"
			raise ValueError(message)

	def calculate_silhouette_coefficient(self, label: str, cluster_label: str, other_label: str):
		""" Calculates the Silhouette Coefficient for a single point.
			Parameters
			----------
			label: str
				THe label of the sample
			cluster_label: str
				The label of the cluster that `label` belongs to.
			other_label: str
				The label of the next closest cluster.
		"""
		pass

	def calculate_silhouette_coefficients(self) -> Dict[str, float]:
		coefficients = dict()
		for cluster_name, cluster in self.genotypes.items():
			if len(cluster) == 1: continue
			closest_cluster = self.get_closest_cluster(cluster_name)
			closest_cluster_members = self[closest_cluster]
			for member in cluster:
				distances_between_other_members = [self.distances[member, other] for other in cluster]
				distances_between_closest_cluster = [self.distances[member, other] for other in closest_cluster_members]
				# -1 sincec `member` is included as 0
				average_distance = sum(distances_between_other_members) / (len(distances_between_other_members) - 1)
				average_distance_closest = sum(distances_between_closest_cluster) / len(distances_between_closest_cluster)
				c = (average_distance_closest - average_distance) / (max(distances_between_closest_cluster + distances_between_other_members))
				coefficients[member] = c
		return coefficients

	@property
	def genotypes(self) -> Dict[str, ClusterType]:
		if self._genotypes is None:
			self._genotypes = self.get_genotypes(self.clusters)
		return self._genotypes

	@staticmethod
	def calculate_within_cluster_variation(cluster: ClusterType, pair_array: DistanceCache) -> float:
		""" Row labels and column names are samples."""

		if len(cluster) > 1:
			cluster_distances = [pair_array[i, j] for i, j in itertools.combinations(cluster, 2)]
		else:
			cluster_distances = [0]
		cluster_mean = statistics.mean(cluster_distances)

		variances = [(i - cluster_mean) ** 2 for i in cluster_distances]
		W = sum(variances) / len(cluster)
		return W

	def calculate_within_cluster_variations(self) -> float:
		""" Calulates the total within-cluster variation for all clusters in `clusters`.
		"""
		cluster_variations = [self.calculate_within_cluster_variation(cluster, self.distances) for cluster in self.clusters]
		return sum(cluster_variations)

	def get_genotypes(self, clusters: List[ClusterType] = None) -> Dict[str, ClusterType]:
		""" Generates labels for each cluster."""
		if clusters is None: clusters = self.clusters
		clusterdict = dict()
		for index, cluster in enumerate(clusters, start = 1):
			# Test if each trajectory shares a common name. Usually due to testing genotypes.
			try:
				_names = {i.split('-')[1] for i in cluster}
			except IndexError:
				_names = set()
			if len(_names) == 1:
				genotype_name = "genotype-" + list(_names)[0]
			else:
				genotype_name = f"genotype-{index}"
			clusterdict[genotype_name] = cluster
		return clusterdict

	def get_closest_cluster(self, label: str) -> str:
		""" Return the label of the closest cluster to `label`"""
		dist = dict()
		for name, cluster in self.genotypes.items():
			if name == label: continue
			d = self.calculate_distance_between_clusters(self[label], self[name])
			dist[name] = d
		return min(dist.items(), key = lambda s: s[1])[0] 



from pathlib import Path
from typing import Dict, Optional, Union

import pandas
from loguru import logger

from muller import dataio, graphics


# noinspection PyUnresolvedReferences
class GGMuller:
	""" Uses the `population` and `edges` table to generate muller plots and timeseries.
		Parameters
		----------
		populations: Union[Path. pandas.DataFrame]
		The population table is essentially a timeseries table in longformat (one observation per row).
		* `Generation`: The timepoint for this row.
		* `Identity`: Name of the series to which this observation belongs.
		* `Population` The measured frequency (as a number in [0,100])
		edges: Union[Path, pandas.DataFrame]
		Describes the lineage of the genotypes.
		* `Identity`: The genotype id.
		* `Parent`: The parent of `Identity`

		Output Files
		------------
		`data/population.tsv': The input population table
		`data/edges.tsv`: The input edges table
		`data/timeseries.tsv`: The condensed vertion of `population.tsv` where
			each column corresponds to a specific timepoint and each row corresponds to a
			single genotype.
		`figures/muller.(png|svg)` The muller diagram.

		Output Structure
		----------------
		folder
		|---- 'data'
		|----|---- 'population.tsv'
		|----|---- `edges.tsv'
		|----|---- `timeseries'
		|----'figures'
		|----|---- 'muller.(png|svg)

	"""

	def __init__(self, render: bool = True, folder: Optional[Path] = None):
		self.render = render
		if folder is None:
			self.paths = dict()
		else:
			self.paths = self.generate_filesystem_structure(folder)

	##############################################################################################################################################
	# --------------------------------------------------------------- Utilities ---------------------------------------------------------------------
	##############################################################################################################################################

	@staticmethod
	def generate_filesystem_structure(parent_folder: Path) -> Dict[str, Path]:
		parent_folder = dataio.checkdir(parent_folder)
		folder_data = dataio.checkdir(parent_folder / "data")
		folder_figure = dataio.checkdir(parent_folder / "figures")
		paths = {
			'outputTablePopulationLong': folder_data / "populations.long.tsv",
			'outputTablePopulationWide': folder_data / "populations.wide.tsv",
			'outputTableEdges':          folder_data / "edges.tsv",
			'outputTableMuller':         folder_data / "mullerformat.tsv",
			'outputFigureMuller':        folder_figure / "mullerplot.png"
		}
		return paths

	@staticmethod
	def load_table(io: Union[str, Path, pandas.DataFrame]) -> pandas.DataFrame:
		if not isinstance(io, pandas.DataFrame):
			# Assume it is one of the other formats
			result = dataio.import_table(io)
		else:
			result = io
		return result

	@staticmethod
	def convert_to_timeseries(table: pandas.DataFrame):
		result = table.pivot(index = 'Identity', columns = 'Generation', values = 'Population')
		return result

	##############################################################################################################################################
	# ------------------------------------------------------------------- main ---------------------------------------------------------------------
	##############################################################################################################################################

	def run(self, populations: Union[str, Path, pandas.DataFrame], edges: Union[str, Path, pandas.DataFrame], output_folder: Optional[Path] = None) -> \
	Dict[str, Path]:
		"""
		Parameters
		----------
		populations: Union[Path. pandas.DataFrame]
		edges: Union[Path, pandas.DataFrame]
		output_folder: Path
			The folder to put the output in. This will override the defaults generated when this workflow was initialized.

		Returns
		-------
		paths: Dict[str,Path]
			The dictionary with the predetermined filenames.
		"""
		if output_folder is not None:
			self.paths = self.generate_filesystem_structure(output_folder)
		# Load the two tables if they ar not already loaded.
		table_populations = self.load_table(populations)
		# The edges table maps genotypes to their corresponding parent.
		# Since this is a simple mapping, convert it to a pandas.Series object
		# 	which makes handling key-value pairs a little more convienient.
		table_edges = self.load_table(edges)
		# Need to save the pandas.Series object as a separate variable so that it isn't changed below.
		series_edges = table_edges.set_index('Identity')['Parent']
		# convert `table_populations` into a wide-format (columns == timepoints, rows = samples) timeseries table.
		table_populations_wide = self.convert_to_timeseries(table_populations)

		# Convert the `populations` and `edges` tables to a muller-df format.
		# The muller_df combines the information from those two tables.
		# The resulting dataframe will have the following columns:
		# Generation	Identity	Population	Frequency	Group_id	Unique_id
		# Generation
		# Identity: The name of the series
		# Groupid: Since each series is plotted twice (once for the bottom portion and again for the top portion)
		# 	whenever there are child genotypes, Each of those two series needs a way of identifying them.
		#	The general format is `identity` + (a|b)
		# Unique_id: Similar to the `Group_id` above, but for each individual timepoint.
		#	ex. `identity` + (a|b) + '_' + `timepoint`

		muller_formatter = dataio.GenerateMullerDataFrame()
		table_muller = muller_formatter.run(series_edges, table_populations)

		# Save the tables before rendering the muller plot in case something goes wrong.
		self.save_table(table_edges, "outputTableEdges")
		self.save_table(table_populations, 'outputTablePopulationLong')
		self.save_table(table_populations_wide, 'outputTablePopulationWide')
		self.save_table(table_muller, 'outputTableMuller')

		filename_figure_muller = self.paths['outputFigureMuller']

		# Generate the muller plot.
		diagram_generator = graphics.MullerPlot(outlines = True, render = True)
		diagram_generator.plot(table_muller, filename_figure_muller)

		return self.paths

	##############################################################################################################################################
	# --------------------------------------------------------------- Output ---------------------------------------------------------------------
	##############################################################################################################################################

	def save_table(self, table, output: Union[str, Path]) -> Path:
		""" Save the current table.
			Parameters
			----------
			table:pandas.DataFrame
				The source table to save. Can be any of the tables used in this workflow, the exact type of which is specified by `output`.
			output: Union{str, Path]
				The path to save the current table as. providing a `Path` object will save the table as that filename.
				If a key from `self.paths` is provided, the filename will lookup the corresponding default from `self.paths,

			Returns
			-------
			filename: Path
				The filename the table was saved as.

		"""
		if isinstance(output, Path):
			# No need to change anything.
			filename = output
		elif isinstance(output, str) and output in self.paths:
			# Should be a key from the `self.paths` dictionary.
			filename = self.paths[output]
		else:
			# `output` is not a Path or a key from `self.keys`.
			valid_keys = sorted(self.paths.keys())
			message = f"Expected a filename object or a key from {valid_keys}. Got '{output}' instead."
			logger.error(message)
			raise ValueError(message)
		table.to_csv(filename, sep = "\t")
		return filename


if __name__ == "__main__":
	pass



import argparse
import itertools
import math
from pathlib import Path
from typing import List, Optional, Union

try:
	from muller import dataio
except ModuleNotFoundError:
	from . import dataio

from dataclasses import dataclass

__VERSION__ = "0.9.0"
DEBUG = False


# For convienience. Helps with autocomplete.
# Not really used, will probably remove later.
@dataclass
class ProgramOptions(argparse.Namespace):
	filename: Path
	output_folder: Path
	sheetname: str = "Sheet1"
	fixed_breakpoint: Optional[float] = None
	detection_breakpoint: float = 0.03
	significant_breakpoint: float = 0.15
	mode: bool = False
	frequencies: List[float] = 0.10
	similarity_breakpoint: float = 0.05
	difference_breakpoint: float = 0.10
	is_genotype: bool = False
	use_filter: bool = True
	annotate_all: bool = False
	save_pvalue: bool = True
	use_strict_filter: bool = False
	method: str = 'matlab'
	metric: str = "similarity"
	known_genotypes: Optional[Path] = None


ACCEPTED_METHODS = ["matlab", "hierarchy", "twostep"]


# noinspection PyTypeChecker
def parse_workflow_options(program_options: argparse.Namespace) -> ProgramOptions:
	"""
		Generates values for each of the program parameters from the given parameters on the command line.
	Parameters
	----------
	program_options

	Returns
	-------

	"""
	#if program_options.name != 'lineage': return program_options
	if program_options.flimit is None:
		program_options.flimit = 1 - program_options.dlimit
	if program_options.known_genotypes:
		program_options.known_genotypes = Path(program_options.known_genotypes)
		starting_genotypes = dataio.import_file.parse_known_genotypes(program_options.known_genotypes)
	else:
		starting_genotypes = None

	program_options.starting_genotypes = starting_genotypes
	if program_options.output_folder is None:
		program_options.output_folder = Path(program_options.filename.parent)

	if program_options.similarity_cutoff is not None and program_options.similarity_cutoff < 0:
		program_options.similarity_cutoff = None

	return program_options


def _parse_frequency_option(frequency: Union[str, List[float]]) -> List[float]:
	if isinstance(frequency, str):
		if ',' in frequency:
			frequencies = list(map(float, frequency.split(',')))
		else:
			frequencies = float(frequency)
	elif isinstance(frequency, float):
		frequencies = frequency
	else:
		frequencies = [float(i) for i in frequency]
	if isinstance(frequencies, float):
		frequencies = [math.fsum(itertools.repeat(frequencies, i)) for i in range(int(1 / frequencies) + 1)]
		frequencies = [round(i, 2) for i in frequencies]
	if 0 not in frequencies:
		frequencies = [0] + frequencies
	frequencies = sorted(frequencies, reverse = True)
	return frequencies


#####################################################################################
################################ Custom Parsers ####################################
#####################################################################################

class FixedBreakpointParser(argparse.Action):
	def __call__(self, parser, namespace, values, option_string = None):
		detected = namespace.dlimit
		if values == "1":
			values = 1 - detected
		else:
			values = float(values)
		setattr(namespace, self.dest, values)


#####################################################################################
############################# Lineage Parser Groups #################################
#####################################################################################
def _create_parser_lineage_group_graphics(parser: argparse.ArgumentParser):
	##############################################################################################################################################
	# -------------------------------------------------------- Graphics Options ------------------------------------------------------------------
	##############################################################################################################################################

	group_graphics = parser.add_argument_group(title = "Graphics Options", description = "Options to refine how the output graphics are generated.")

	group_graphics.add_argument(
		"--annotate-all",
		help = "Adds all gene labels to the muller plots, instead of the top three.",
		action = "store_true",
		dest = "annotate_all"
	)
	group_graphics.add_argument(
		"--genotype-colors",
		help = "An optional map of genotypes to specified colors.",
		action = "store",
		type = Path,
		default = None,
		dest = "genotype_palette_filename"
	)
	group_graphics.add_argument(
		"--no-render",
		help = "The svg files will not be generated.",
		action = "store_false",
		dest = "render"
	)
	group_graphics.add_argument(
		"--no-outline",
		help = 'Disables the white outline in the muller plots.',
		action = 'store_false',
		dest = 'draw_outline'
	)
	group_graphics.add_argument(
		"--highlight",
		help = "A comma-separated list of genotype names or annotations to highlight in the generated graphics.",
		action = "store",
		dest = "highlight"
	)
	group_graphics.add_argument(
		"--highlight-color",
		help = "What the color of highlighted genotypes should be. Only HEX color codes are supported.",
		action = 'store',
		dest = 'highlight_color',
		default = "#FFFF00"
	)

	group_graphics.add_argument(
		"--no-smoothing",
		help = "Whether to smooth the values represented in the timeseries plots.",
		action = 'store_false',
		dest = "smooth_plot"
	)

	return group_graphics


# noinspection PyTypeChecker,PyTypeChecker,PyTypeChecker
def _create_parser_lineage_group_data(parser: argparse.ArgumentParser):
	##############################################################################################################################################
	# ----------------------------------------------------- Additional Input Files ---------------------------------------------------------------
	##############################################################################################################################################

	group_data = parser.add_argument_group(title = "Input Data", description = "Additional data that can be used to improve the analysis.")
	group_data.add_argument(
		"--filename-pairwise",
		help = "Path to a table with pairwise distance calculations from a previous run using identical input parameters. Should be located " \
			   "in `tables/.distances.tsv in the output folder generated from the previous run. This table will be used rather than re-calculating " \
			   "all the pairwise distances again which may take a long time for very large datasets.",
		action = "store",
		dest = "filename_pairwise",
		type = Path,
		default = None
	)

	group_data.add_argument(
		"--gene-alias",
		help = "An optional two-column file with more accurate gene names. This is usefull when using a reference annotated via prokka.",
		action = "store",
		type = Path,
		default = None,
		dest = 'alias_filename'
	)
	group_data.add_argument(
		"-g", "--known-genotypes",
		help = "A file with trajectories known to be in the same genotypes. "
			   "Each genotype is defined by a comma-delimited line with the labels of the member trajectories.",
		action = "store",
		default = None,
		dest = "known_genotypes"
	)
	group_data.add_argument(
		"--known-ancestry",
		help = "A file designating the known ancestry of certain genotypes. Formatted like the ggmuller edges table.",
		dest = 'known_ancestry',
		default = None,
		type = Path
	)


def _create_parser_lineage_group_genotype_generation(parser: argparse.ArgumentParser):
	##############################################################################################################################################
	# ------------------------------------------------------ General Analysis Options ------------------------------------------------------------
	##############################################################################################################################################
	analysis_group = parser.add_argument_group(title = "Genotype and Lineage Parameters")
	analysis_group.add_argument(
		"--threads",
		help = "The number of processes to use. Adding more threads than available cpu cores provides no speedup.",
		action = "store",
		dest = "threads",
		type = int,
		default = 1
	)

	analysis_group.add_argument(
		"--metric",
		help = "The distance metric to use when clustering mutaitons into genotypes.",
		action = "store",
		dest = "metric",
		type = str,
		default = "binomial"
	)
	analysis_group.add_argument(
		"--similarity-cutoff",
		help = "Used to group trajectories into genotypes. The 10th percentile of the trajectory distances is used if None.",
		action = "store",
		dest = "similarity_cutoff",
		type = float,
		default = None
	)
	analysis_group.add_argument(
		"-p", "--pvalue",
		help = "The p-value to use for the statistics tests",
		action = "store",
		dest = "pvalue",
		type = float,
		default = 0.05
	)

	analysis_group.add_argument(
		'--fixed',
		help = "The minimum frequency at which to consider a mutation fixed.",
		action = FixedBreakpointParser,
		dest = 'flimit',
		type = float,
		default = None
	)
	analysis_group.add_argument(
		"-d", "--detection",
		help = "The uncertainty to apply when performing frequency-based calculations. \
			For example, a frequency at a given timepoint is considered undetected if it falls below 0 + `uncertainty`.",
		action = 'store',
		default = 0.03,
		dest = 'dlimit',
		type = float
	)
	analysis_group.add_argument(
		"-s", "--significant",
		help = "The frequency at which to consider a genotype significantly greater than zero.",
		action = 'store',
		default = 0.15,
		dest = "slimit",
		type = float
	)
	analysis_group.add_argument(
		"--covariance",
		help = "Controls how much a nested and unnested genotype should be correlated/anticorrelated to be considered significant",
		default = 0.01,
		dest = "derivative_cutoff",
		type = float
	)
	analysis_group.add_argument(
		"--liberal",
		help = "Whether to allow genotypes to be placed under more recent nested genotypes when an earlier genotype also passes the score check.",
		action = "store_false",
		dest = "conservative"
	)

	return analysis_group


def _create_parser_lineage_group_filter(parser: argparse.ArgumentParser):
	##############################################################################################################################################
	# --------------------------------------------------- Genotype Filtering Options -------------------------------------------------------------
	##############################################################################################################################################
	group_filter = parser.add_argument_group(title = "Filtering Parameters", description = "Parameters to control the filtering process.")
	group_filter.add_argument(
		"--disable-all-filters",
		help = "Disables genotype filtering.",
		action = 'store_false',
		dest = 'use_filter'
	)
	group_filter.add_argument(
		"--strict-filter",
		help = """By default, the filters allow trajectories to appear both before and after a genotype"""
			   """fixes as long as they were undetected at the timepoint the sweep occurs. This generally"""
			   """represents mutations which appear, are removed during a genotype sweep, and reappear """
			   """afterwards. Using `--strict-filter` would remove these trajectories.""",
		action = "store_true",
		dest = "use_strict_filter"
	)
	group_filter.add_argument(
		"--disable-filter-single",
		help = "Whether to remove trajectories which only exist at a single timepoint.",
		action = "store_false",
		dest = "use_filter_single"
	)
	group_filter.add_argument(
		"--disable-filter-startsfixed",
		help = "Whether to remove mutations which are 'fixed' at the first timepoint.",
		action = "store_false",
		dest = "use_filter_startsfixed"
	)
	group_filter.add_argument(
		"--filter-constant",
		help = "Sets the delta value by which a mutational trajectory must vary by to not be removed for being a constant mutation. Set to 0 to disable",
		action = "store",
		dest = "filter_constant",
		default = 0.1,
		type = float
	)
	group_filter.add_argument(
		"--disable-genotype-filter",
		help = "Disables filtering based on genotypes.",
		action = "store_false",
		dest = "use_filter_genotype"
	)

	return group_filter


# noinspection PyTypeChecker,PyTypeCheckerFalse
def _create_parser_lineage_group_main(parser: argparse.ArgumentParser):
	group_main = parser.add_argument_group(title = "Main Options")

	group_main.add_argument(
		'-i', '--input',
		help = "The table of trajectories to cluster.",
		action = 'store',
		dest = 'filename',
		type = Path,
		required = True
	)
	group_main.add_argument(
		'-o', '--output',
		help = "The folder to save the files to.",
		action = 'store',
		dest = 'output_folder',
		type = Path,
		default = None
	)
	group_main.add_argument(
		"--name",
		help = 'Prefix to use when naming the output files. defaults to the dataset filename.',
		action = 'store',
		type = str,
		dest = 'name',
		default = None
	)
	##############################################################################################################################################
	# --------------------------------------------------------- Input Data Options ---------------------------------------------------------------
	##############################################################################################################################################
	# input_group = parser.add_argument_group(title = "Input Dataset Options", description = "Options related to the input dataset.")

	group_main.add_argument(
		"--sheetname",
		help = "Indicates the sheet to use if the input table is an excel workbook and the data is not in Sheet1",
		action = 'store',
		dest = 'sheetname',
		default = 0
	)
	group_main.add_argument(
		"--genotypes", "--cohorts",
		help = "Indicates that the input table contains genotypes rather than trajectories.",
		action = 'store_true',
		dest = 'is_genotype'
	)


	return group_main


#####################################################################################
########################## Main Application Parsers #################################
#####################################################################################
def create_lineageplot_parser(subparsers):
	parser = subparsers.add_parser('lineageplot')
	group_main = parser.add_argument_group(title = "Lineageplot Options")
	group_main.add_argument(
		"edges",
		help = "Path to an `edges` table with both `identity` and `parent` columns. An optional `annotations` column"
			   "may also be included.",
		type = Path
	)
	group_main.add_argument(
		"filename",
		help = "Name of where to save the plot.",
		type = Path
	)
	group_main.add_argument(
		"--sheetname",
		help = "The sheet to use if the input file is an excel sheet.",
		type = str,
		required = False
	)
	return group_main


def create_lineage_parser(subparsers) -> argparse.ArgumentParser:
	parser = subparsers.add_parser("lineage")

	# Broke up the indivisual groups into their own functions because this function was large enough to make browsing it annoying.
	_create_parser_lineage_group_main(parser)
	_create_parser_lineage_group_data(parser)
	_create_parser_lineage_group_genotype_generation(parser)
	# _create_parser_group_filter(parser)
	_create_parser_lineage_group_graphics(parser)

	return parser


# noinspection PyTypeChecker,PyTypeChecker
def create_benchmark_parser(subparsers) -> argparse.ArgumentParser:
	""" Defines options for utilities that complement the scripts.
		TODO: Add a script to directly convert breseq output to muller input.
	"""
	parser_benchmark: argparse.ArgumentParser = subparsers.add_parser(
		"benchmark",
		help = "Runs a series of benchmarks to test the time required to calculate all pairwise distances of a dataset based on available processes."
	)
	parser_benchmark.add_argument(
		"--dataset",
		help = "The location of a test dataset to use. Should be large enough that adding additional processes will actually imrpove parsing speed.",
		type = Path
	)
	parser_benchmark.add_argument(
		"--output",
		help = "The filenme of the output graph. Should be a .png file.",
		type = Path
	)

	return parser_benchmark


# noinspection PyTypeChecker,PyTypeChecker,PyTypeChecker
def create_muller_parser(subparsers) -> argparse.ArgumentParser:
	parser_muller: argparse.ArgumentParser = subparsers.add_parser(
		"muller",
		help = "Generates muller plots from a pair of `population` and `edges` tables."
	)

	parser_muller.add_argument(
		"population",
		help = "Path to the population table.",
		type = Path
	)

	parser_muller.add_argument(
		"edges",
		help = "Path to the edges table delineating the lineage of the population genotypes.",
		type = Path
	)
	parser_muller.add_argument(
		"output",
		help = "Path to save the generated muller plot as.",
		type = Path
	)

	return parser_muller


def create_timeseries_parser(subparsers) -> argparse.ArgumentParser:
	parser_timeseries: argparse.ArgumentParser = subparsers.add_parser(
		"timeseries",
		help = "A small utility to plot timeseries tables."
	)
	parser_timeseries.add_argument(
		"timeseries",
		help = "A table with either a `Trajectory` column or a `Genotype` column.",
		type = Path
	)
	return parser_timeseries


def create_parser() -> argparse.ArgumentParser:
	parser_parent = argparse.ArgumentParser(
		description = "Infers genotypes and lineage based on mutation frequency timeseries data observed over the course of an evolutionary biology experiment.",
		formatter_class = argparse.ArgumentDefaultsHelpFormatter
	)
	parser_parent.add_argument(
		"-v", "--version",
		action = 'version',
		version = f"%(prog)s {__VERSION__}"
	)

	subparsers = parser_parent.add_subparsers(dest = 'name')  # Each subparser can be identifies by the `name` attribute.
	create_lineage_parser(subparsers)
	create_benchmark_parser(subparsers)
	create_muller_parser(subparsers)
	create_lineageplot_parser(subparsers)

	return parser_parent


def get_arguments(arguments: Optional[List[str]] = None) -> argparse.Namespace:
	""" Implemented here to make sure the default parameters are properly applied. """
	parser = create_parser()
	args = parser.parse_args(arguments)

	args = parse_workflow_options(args)
	return args




from typing import Dict, List, Tuple

import pandas


def get_child_nodes(tree: pandas.DataFrame, label: str) -> List[str]:
	"""Retrieves all child node for the given label from the tree. Includes `label` in the output"""

	children = list()
	for index in tree.index:
		parents = get_parent_nodes(tree, index)
		if label in parents:
			children.append(index)
	return children


def get_parent_nodes(tree: pandas.DataFrame, label: str) -> List[str]:
	"""Retrieves all parent nodes for the given node."""
	parents = []
	while label != 'genotype-0':
		try:
			parent = tree.loc[label]['Parent']
		except KeyError:
			parent = 'genotype-0'
		parents.append(parent)
		label = parent
	return parents


def parse_tree(edges: pandas.Series) -> pandas.DataFrame:
	"""
		Determines the clade and distance from root of every leaf and node in the ggmuller edges table.
	Parameters
	----------
	edges

	Returns
	-------
	pandas.Series
		- index: str
			The leaf/node genotype label
		- columns
			- 'clade': str
				The root genotype that forms the base of the clade.
			- ''distance': int
				The distance from the node/leaf to the root genotype.
	"""
	edges = edges.copy(deep = True)  # To prevent unintended alterations

	#lineage_table = edges.set_index('Identity')['Parent']
	lineage_table = edges

	clades, iterations = zip(*[determine_clade(lineage_table, i) for i in lineage_table.index])
	leaf_table = lineage_table.to_frame().reset_index()
	leaf_table['clade'] = clades
	leaf_table['iterations'] = iterations
	leaf_table = leaf_table.set_index('Identity')
	return leaf_table.sort_values(by = ['clade', 'iterations'])


def determine_clade(parents: pandas.Series, label: str) -> Tuple[str, int]:
	iteration = 0
	while True:
		iteration += 1
		index = parents[label]
		if index != 'genotype-0':
			label = index
		else:
			break
	return label, iteration


def common_elements(left, right) -> int:
	""" Returns the number of elements that are common between `left` and `right`"""
	common = set(left) & set(right)
	return len(common)


def tokenize(elements: List[str]):
	""" Converts a list of strings into a list of unique words found in all the string elements."""
	string = " ".join(elements)
	tokens = string.split(" ")
	return [i.strip() for i in tokens if i]


def group_clades(clade_annotations: Dict[str, List[str]]) -> List[List[str]]:
	"""
		Groups clades by similarity.
		Example
		-------
		clades = {
			'genotype-10': ['dltB '],
			'genotype-11': ['dltB Q111*'],
			'genotype-16': ['PROKKA_00173|PROKKA_00174 '],
			'genotype-6': ['SPAR113_1988 G119C'],
			'genotype-9': ['dltB ']
    	}
    	result = group_clades(clades)
    	result == [['genotype-10', 'genotype-11', 'genotype-9'], ['genotype-16'], ['genotype-6']]
	"""
	tokens = {k: tokenize(v) for k, v in clade_annotations.items()}

	result = list()
	seen = set()
	for key, value in tokens.items():
		if key in seen: continue
		seen.add(key)
		related = [k for k, v in tokens.items() if common_elements(value, v)]
		result.append(related)
		seen = seen | set(related)

	return result



import math
from typing import Dict, List, Mapping, Optional, Tuple, Union

import pandas


class Ancestry:
	""" Holds the possible ancestry candidates as well as the confidance score for each.
		Parameters
		----------
		initial_background: pandas.Series
			The genotype which reached the maximum relative frequency within the population. Serves as the first
			nested genotype to compare all others against.
		timepoints: pandas.DataFrame
			A copy of the genotype timeseries table. Currently only used in the `self.get_sum_of_backgrounds` method.
			May be worth removing later to reduce the number of dependent parameters.
		cautious: bool = True
			Indicates whether to favor the oldest genotype within at least 2 points of the maximum genotype.
			Basically controlls the likliness that these scripts will assign a genotype to an olser lineage
			rather than nesting the genotype under a newer lineage.
	"""

	def __init__(self, initial_background: pandas.Series, timepoints: pandas.DataFrame, cautious: bool = True):
		self.cautious = cautious
		self.initial_background_label: str = initial_background.name

		# The minimum score to consider a genotype as a possible ancestor.
		#self.minimum_score = 1 if not self.cautious else 0.01 # basically anything non-zero.
		self.minimum_score = 1
		# The number of points separating a genotype from the maximum scored genotype
		# by which an older genotype will be considered as a viable newest ancestor
		# Make a copy to prevent unintended modifications to source table.
		self.score_window = 2
		self.timepoints = timepoints.copy()

		# Keep track of the parent and confidence for each new genotype.
		self.confidence: Dict[str, List[Tuple[str, float]]] = dict()
		self.ancestral_genotype = 'genotype-0'

		self.nests: Dict[str, List[str]] = dict()
		self.add_genotype_to_background(initial_background.name, self.ancestral_genotype, 1)

	def add_genotype_to_background(self, unnested_label: str, nested_label: str, priority: Union[int, float]) -> None:
		if unnested_label not in self.nests:
			self.nests[unnested_label] = list()

		self.nests[unnested_label].append(nested_label)

		if unnested_label not in self.confidence:
			self.confidence[unnested_label] = []
		self.confidence[unnested_label].append((nested_label, priority))

	def get(self, label: str) -> List[str]:
		return self.nests[label]

	def is_a_member(self, label: str) -> bool:
		return label in self.nests.keys()

	def get_sum_of_backgrounds(self) -> pandas.Series:
		background_labels = [k for k in self.nests.keys() if self.is_a_background(k)]
		background_frequencies = self.timepoints.loc[background_labels]
		total = background_frequencies.sum()
		return total

	def is_a_background(self, element: str) -> bool:
		background = self.get(element)
		return len(background) == 1 or (len(background) == 2 and 'genotype-0' in background)

	def get_highest_priority_legacy(self, label: str) -> Tuple[Optional[str], float]:
		candidates = self.confidence.get(label, [])
		if candidates:
			# Explicity tell the sorting method to use the priority score.
			# This will prevent the method from using the genotype name to sort the elements,
			# So ties should be broken by whichever candidate was added as a candidate first.
			candidate, score = max(candidates, key = lambda s: s[1])

		else:
			# `candidates` was an empty sequence.
			candidate = None
			score = math.nan

		if score < 1:
			candidate = None

		return candidate, score

	def get_highest_priority(self, label: str) -> Tuple[Optional[str], float]:
		""" Returns the genotype label representing the newest ancestor for the genotype indicated by `label`."""
		candidates = self.confidence.get(label, [])
		maximum_genotype, maximum_score = max(candidates, key = lambda s: s[1])

		for candidate, score in candidates:
			# Make sure the score is within 2 or so of the maximum.
			# Will output the genotype with the maximum score if no other genotype exists with 2 points of the maximum.
			if score > self.minimum_score and abs(maximum_score - score) <= self.score_window:
				return candidate, score
		else:
			return None, math.nan


	def as_ancestry_table(self) -> pandas.Series:
		table = list()
		for identity, background in self.nests.items():
			# parent = self.get_highest_priority(identity)
			# if parent is None:
			parent, score = self.get_highest_priority(identity)
			if parent == identity or parent is None:
				parent = self.ancestral_genotype
			row = {
				'Parent':   parent,
				'Identity': identity
			}
			table.append(row)
		table = pandas.DataFrame(table)[['Parent', 'Identity']]  # Reorder columns

		return table.set_index('Identity')['Parent']

	def as_dict(self) -> Mapping[str, str]:
		return self.as_ancestry_table().to_dict()

	def priority_table(self) -> pandas.DataFrame:
		data = list()
		for identity, background in self.nests.items():
			# parent = self.get_highest_priority(identity)
			# if parent is None:
			parent, score = self.get_highest_priority(identity)
			if parent == identity or parent is None:
				parent = self.ancestral_genotype
			row = {
				'parent':   parent,
				'identity': identity,
				'score':    score
			}
			data.append(row)
		return pandas.DataFrame(data)

	def to_table(self) -> pandas.DataFrame:
		data = list()
		for identity, candidates in self.confidence.items():
			for candidate, score in candidates:
				row = {
					'identity':  identity,
					'candidate': candidate,
					'score':     score
				}
				data.append(row)
		return pandas.DataFrame(data)



from pathlib import Path

import pandas


class OutputStructure:
	def __init__(self, folder: Path):
		self.filename_genotypes = folder / "tables" / "lineage.trajectory.genotypes.tsv"
		self.filename_edges = folder / "tables" / "lineage.trajectory.lineage.edges.tsv"
		self.filename_figure_mullerpanel = folder / "lineage" / "lineage.trajectory.mullerpanel.png"
		self.filename_muller_table = folder / "tables" / "lineage.trajectory.lineage.muller.tsv"

		self.table_genotypes = pandas.read_csv(self.filename_genotypes, sep = "\t")
		self.table_edges = pandas.read_csv(self.filename_edges, sep = "\t")
		self.table_muller = pandas.read_csv(self.filename_muller_table, sep = "\t")
		# TODO make sure the genotypes table is saved with a Genotype column.


import pandas

try:
	from muller.inheritance import timepoint_detection
except ModuleNotFoundError:
	from muller.inheritance import timepoint_detection


class SortGenotypeTableWorkflow:
	"""
		Sorts the muller_genotypes based on when they were first detected and first fixed.
		dlimit, slimit, flimit: float
			The three breakpoints that are used to determine the sort order of the genotypes.
		breakpoints: List[float]
			Frequencies with which to group genotypes into sortable bins. Each bin will be sorted individually then added to the final output.
	"""

	def __init__(self, dlimit: float, flimit: float):
		self.dlimit = dlimit
		self.slimit = 0.15
		self.flimit = flimit
		# The `breakpoints` value is a bit arbitrary, so it should be safe to hard-code it.
		# 	This will actually prevent the most common error when sorting genotypes (i.e. no breakpoints given) so it's worth
		#	hard-coding it to prevent that issue.
		# TODO: Make sure this is hard-coded globally.
		# self.breakpoints = sorted([1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.0] + [self.dlimit, self.flimit], reverse = True)
		self.breakpoints = [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.0]

	@staticmethod
	def _build_sorted_frequency_table(original_frequencies: pandas.DataFrame, thresholds: pandas.DataFrame) -> pandas.DataFrame:
		# Sort genotypes based on frequency if two or more share the same fixed timepoint, detection timpoint, threshold timepoint.
		# Iterate over the conbinations of 'firstFixed', 'firstDetected', and 'firstThreshold' and sort trajectories that belong to the sample combination.
		# Need to build a new trajectory table based on the sorted timepoint table.

		freq_groups = list()
		# Group the
		groups = thresholds.groupby(by = list(thresholds.columns))
		for (ff, fd, ft), group in groups:
			# Each group will look like this:
			#   	firstFixed firstDetected firstThreshold
			# 15          0            25              0
			# 10          0            25              0
			# if ft == 130:  # dummy value assigned above. Revert back to 0 since '130' isn't a valid timepoint.
			#	ft = 0
			# Get a table of the original frequencies at each timepoint for the genotypes present in `group`
			trajectories: pandas.DataFrame = original_frequencies.loc[group.index]
			if len(group) < 2:
				# There is only one genotype in this combination of timepoints. No need to sort.
				freq_groups.append(trajectories)
			else:
				# More than one genotype share this combination of key timepoints. Sort by frequency.
				# Sort from highest to lowest using the timpoint columns as the sorting keys.
				# trajectories = trajectories.sort_values(by = [ff, ft, fd], ascending = False)
				trajectories = trajectories.sort_values(by = [fd, ft, ff], ascending = False)
				freq_groups.append(trajectories)

		try:
			freq_df = pandas.concat(freq_groups)
		except ValueError:
			# Can't concatenate an empty dataframe
			freq_df = None

		return freq_df

	def _sort_genotype_frequencies(self, genotype_trajectories: pandas.DataFrame, frequency_breakpoint: float) -> pandas.DataFrame:
		"""
			Sorts the genotype table based on the timepoints that first satisfy the threshold requirements.
		Parameters
		----------
		genotype_trajectories
		frequency_breakpoint
		"""
		# Get the values for each genotype at the first timepoint. This will be used to check if the genotype existed at the first timepoint
		# or has an initial detected timepoint of 0 due to the way pandas returns indicies.
		detection_cutoff = min(self.dlimit, frequency_breakpoint)
		transposed = genotype_trajectories.transpose()

		first_detected_reduced = timepoint_detection.get_first_detected_timepoint(transposed, detection_cutoff)
		first_above_threshold_reduced = timepoint_detection.get_first_significant_timepoint(transposed, self.slimit)
		# Use the frequency breakpoint rather than the fixed cutoff.
		first_fixed_reduced = timepoint_detection.get_first_fixed_timepoint(transposed, frequency_breakpoint)

		combined_df: pandas.DataFrame = pandas.concat(
			[first_fixed_reduced, first_detected_reduced, first_above_threshold_reduced],
			axis = 1, sort = False)
		df = combined_df[~combined_df['firstFixed'].isna()]

		if frequency_breakpoint == self.flimit:
			df = df.dropna()
		df = df[['firstDetected', 'firstFixed', 'firstThreshold']]
		sorted_frequencies = df.sort_values(by = ['firstDetected', 'firstThreshold'], ascending = True)

		# Sort genotypes based on frequency if two or more share the same fixed timepoint, detection timpoint, threshold timepoint.
		# Iterate over the conbinations of 'firstFixed', 'firstDetected', and 'firstThreshold' and sort trajectories that belong to the sample combination.
		# Need to build a new trajectory table based on the sorted timepoint table.
		freq_df = self._build_sorted_frequency_table(genotype_trajectories, sorted_frequencies)

		return freq_df

	@staticmethod
	def remove_trajectories_below_threshold(df: pandas.DataFrame, threshold: float) -> pandas.DataFrame:
		""" Removes all genotypes that do not have at least one frequency above the given threshold."""
		result = df[df.max(axis = 1) >= threshold]
		return result

	def run(self, unsorted_genotypes: pandas.DataFrame):
		"""
			Sorts the muller_genotypes based on when they were first detected and first fixed.
		Parameters
		----------
		unsorted_genotypes: pandas.Dataframe
			A dataframe with the mean frequency of each genotype, derived from the member trajectories
			in that genotype. Each row should correspond to a single genotype.
		Returns
		-------
		sorted_genotype: pandas.DataFrame
		"""
		sorted_genotypes = list()
		current_genotypes: pandas.DataFrame = unsorted_genotypes.copy()
		for frequency in self.breakpoints:
			# Ignore genotypes that do not have at least on timepoint exceeding the current frequency.
			# This should be taken care of suring the filtering step.
			genotypes_above_threshold = self.remove_trajectories_below_threshold(current_genotypes, frequency)

			sorted_dataframe = self._sort_genotype_frequencies(
				genotype_trajectories = genotypes_above_threshold,
				frequency_breakpoint = frequency
			)
			if sorted_dataframe is not None:
				current_genotypes = current_genotypes.drop(sorted_dataframe.index)
				sorted_genotypes.append(sorted_dataframe)
		df = pandas.concat(sorted_genotypes, sort = False)
		df = unsorted_genotypes.reindex(df.index)

		# Make sure the genotype column is labelled correctly.
		df.index.name = "Genotype"

		return df



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import pandas


@dataclass
class DataWorkflowBasic:
	# Used to organize the output from the workflow.DistanceCache
	# Should only cover relevant data about the scripts as a whole.
	version: str  # The version of the scripts
	filename: Path  # The filename of the input dataset.
	program_options: Any # The command-line arguments passed to the scripts.
	def save(self, folder:Path):
		options = vars(self.program_options)
		options['filename'] = self.filename
		options['version'] = self.version

		file_name = folder / "options.json"
		# Need to conver `Path` to `str`
		options = {key: (str(value)) for key, value in options.items()}
		file_name.write_text(json.dumps(options, sort_keys = True, indent = 4))

@dataclass
class DataGenotypeInference:
	"""
		Contains data related to trajectories and genotypes.

	"""
	# Include the original trajectory table.
	table_trajectories: pandas.DataFrame
	table_trajectories_info: Optional[pandas.DataFrame]
	# A table with inferred genotypes sorted by maximum frequency and earliest timepoint.
	# Each genotype is the mean of consituent trajectories.
	table_genotypes: pandas.DataFrame
	# Map of genotypes to member trajectories.
	genotype_members: Dict[str,List[str]]

	# A pairwise distance matrix showing the calculated distance between an given pair of trajectories.
	# Type: 'muller.clustering.metrics.distance_cache.DistanceCache'
	matrix_distance: Optional[Any]
	# Generated from the hierarchal clustering step. Links trajectories based on the pairwise distance.
	clusterdata: Optional["DataHierarchalCluster"]

	def save(self, folder:Path, prefix:str):
		"""
			Saves the data generated while infering genotypes.
			.folder
			|---- {prefix}.genotypes.tsv
			|---- {prefix}.trajectories.tsv
			|---- {prefix}.distancematrix.tsv
			|---- {prefix}.linkagetable.tsv
		"""
		delimiter = '\t'
		suffix = 'tsv'
		filename_table_trajectory = folder / (prefix + f'.trajectories.{suffix}')
		filename_table_genotypes = folder / (prefix + f'.genotypes.{suffix}')
		filename_table_distance_matrix = folder / (prefix + f'.distancematrix.{suffix}')
		filename_table_linkage_matrix = folder / (prefix + f'.linkagetable.{suffix}')

		# Include the `genotype_members` table in the `genotypes` table.
		# Convert the `genotype_members` from Dict[str,List[str]] to Dict[str,str]
		members = {key:'|'.join(values) for key, values in self.genotype_members.items()}
		# Add the genotype members to the `genotypes` table.
		self.table_genotypes['members'] = members

		# Save the data
		self.table_trajectories.to_csv(filename_table_trajectory, sep = delimiter)
		self.table_genotypes.to_csv(filename_table_genotypes, sep = delimiter)
		if self.clusterdata is not None:
			self.clusterdata.table_linkage.to_csv(filename_table_linkage_matrix, sep = delimiter)
		if self.matrix_distance is not None:
			self.matrix_distance.squareform().to_csv(filename_table_distance_matrix, sep = delimiter)

		# Need to remove the `members` column from the genotype table so that the graphics workflow uses a purely numeric table
		self.table_genotypes.pop('members')

@dataclass
class DataHierarchalCluster:
	""" Output from the hierarchal cluster. """
	clusters: List[List[str]]
	table_linkage: Optional[pandas.DataFrame]
	distance_cutoff: float
	distance_quantile: float
	def to_dict(self)->Dict[str,Any]:
		data = {
			'clusters': self.clusters,
			'distanceCutoff': self.distance_cutoff,
			'distanceQuantile': self.distance_quantile
		}

		return data


@dataclass
class DataGGmuller:
	""" Holds data related to the current implementation of ggmuller."""
	table_populations: pandas.DataFrame
	series_edges: pandas.Series
	table_muller: pandas.DataFrame

	def save(self, folder:Path, prefix: str):
		"""
			.
			|---- .ggmuller.population.tsv
			|---- .ggmuller.edges.tsv
			|---- .script.r
		"""
		delimiter = '\t'
		suffix = 'tsv'

		filename_table_populations = folder / (prefix + f'.ggmuller.population.{suffix}')
		filename_table_edges = folder / (prefix + f'.ggmuller.edges.{suffix}')
		filename_script_r = folder / (prefix + f'.script.r')

		self.table_populations.to_csv(filename_table_populations, sep = delimiter)
		self.table_edges.to_csv(filename_table_edges, sep = delimiter)
		filename_script_r.write_text(self.script_r)

@dataclass
class DataGenotypeLineage:
	""" Holds variables generated during the lineage inference step."""
	# The resulting scores of each unnested genotype to cancidate nested genotypes.
	table_scores: pandas.DataFrame
	clusters: Any # muller.inheritance.genotype_ancestry.Ancestry

	def save(self, folder:Path, prefix:str):
		delimiter = "\t"
		suffix = "tsv"

		filename_table_scores = folder / (prefix + f'.lineage.scores.{suffix}')
		filename_table_populations = folder / (prefix + f".lineage.populations.{suffix}")
		filename_table_edges = folder / (prefix + f".lineage.edges.{suffix}")
		filename_table_muller = folder / (prefix + f".lineage.muller.{suffix}")

		self.table_scores.to_csv(filename_table_scores, sep = delimiter)
		self.table_populations.to_csv(filename_table_populations, sep = delimiter)
		self.table_edges.to_csv(filename_table_edges, sep = delimiter, header = True) # Added header option to avoid the FutureWarning
		self.table_muller.to_csv(filename_table_muller, sep = delimiter, index = False)


if __name__ == "__main__":
	pass


from typing import *

import matplotlib.pyplot as plt
import pandas


def generate_lineage_table(genotypes:pandas.DataFrame, lineages:Dict[str,str]):


	groups = genotypes.groupby(by = lambda s: lineages[s])
	lineage_table = list()
	for lineage, group in groups:
		average = group.max()
		lineage_table.append(average)
	return pandas.DataFrame(lineage_table)


def calculate_score(table:pandas.DataFrame, flimit:float = 0.97)->float:
	ancestor = get_ancestor_series(table, 0.97)
	genotypes = table[[i for i in table.columns if i not in ancestor.index]]
	frequencies = genotypes.sum()

	quad_frequencies = frequencies**2
	score = quad_frequencies.sum()/len(table.columns)
	print(frequencies)
	print(quad_frequencies)
	print(score)


def check_lineage(genotypes:pandas.DataFrame, lineages:pandas.Series):

	ancestor = get_ancestor_series(genotypes, 0.97)
	genotypes = genotypes[[i for i in genotypes.columns if i not in ancestor.index]]
	frequencies = genotypes.sum()


	fig, ax = plt.subplots(figsize = (10,10))

	ax.plot(frequencies)

	plt.show()


def get_lineage(edges:pandas.Series, identity:str):
	while True:
		parent = edges.loc[identity].iloc[0]
		if parent != 'genotype-0':
			identity = parent
		else:
			break
	return identity

def get_first_nonancestor_timepoint(table:pandas.DataFrame, flimit:float = 0.97)->int:
	""" Returns the first timepoint that the ancestor was not present in."""
	for column_label in table.columns:

		column = table[column_label]
		if column.sum() >= flimit:
			return column_label
	else:
		# If the ancestor was never eradicated, return the last timepoint.
		return column_label

def get_ancestor_series(table:pandas.DataFrame, flimit:float = 0.97)->pandas.Series:

	endpoint = get_first_nonancestor_timepoint(table, flimit)
	early_table = table[[i for i in table.columns if i < endpoint]]
	nonancestor = early_table.sum()
	ancestor = 1 - nonancestor

	return ancestor


def read_tables()->Tuple[pandas.DataFrame, pandas.DataFrame]:
	filename = "/home/cld100/Documents/github/muller_diagrams/manuscript_scripts/03 - similarity cutoff comparison/results/sc-1/real.nature12344-s2.BYB1-G07.genotypes.tsv"
	filename_edges = "/home/cld100/Documents/github/muller_diagrams/manuscript_scripts/03 - similarity cutoff comparison/results/sc-1/tables/real.nature12344-s2.BYB1-G07.edges.tsv"
	table = pandas.read_csv(filename, sep = "\t")
	table = table.set_index('Genotype')
	if 'members' in table.columns:
		table.pop('members')
	edges = pandas.read_csv(filename_edges, sep = "\t")
	edges = edges.set_index('Identity')

	table.columns = [int(i) for i in table.columns]

	return table, edges

def main():
	table_genotypes, edges = read_tables()
	lineages = dict()
	for identity in edges.index:
		lineage = get_lineage(edges, identity)
		lineages[identity] = lineage

	lineage_table = generate_lineage_table(table_genotypes, lineages)
	#table_genotypes['lineage'] = [get_lineage(edges, i) for i in table_genotypes.index]
	calculate_score(lineage_table)

if __name__ == "__main__":
	main()


"""
	Benchmarks the distance calculator. Basically provides an idea of how long it will take to calculate the distance across all pairwise
	trajectories based on how many trajectories there are and how many timepoints were sampled.
"""
import time
from pathlib import Path
from typing import *

import matplotlib.pyplot as plt

plt.style.use("fivethirtyeight")
import pandas
from loguru import logger

from muller import dataio, widgets
from muller.clustering import DistanceCalculator


def load_datasets(filenames: List[Path]) -> List[pandas.DataFrame]:
	tables = list()
	for filename in filenames:
		logger.debug(f"Loading {filename}...")
		t = dataio.import_table(filename, sheet_name = 'trajectory', index = "Trajectory")

		# Remove any unnecessary columns so that the entire table is numeric
		t = t[widgets.get_numeric_columns(t.columns)]

		tables.append(t)
	return tables


def load_highresolution_datasets(filename:Path) -> List[pandas.DataFrame]:

	# Basically want to use this large dataset to generate tables of various sizes.
	table = dataio.import_table(filename, index = "Trajectory")
	table = table[widgets.get_numeric_columns(table.columns)]
	table.columns = list(range(len(table.columns)))
	dfs = list()
	for i in range(2, 10):
		size = int(2 ** i)
		df = table.iloc[:size]
		dfs.append(df)

	return dfs


def get_datasets() -> List[Path]:
	# Use the same datasets as `tests`. May also want to link some larger datasets manually.
	folder_data = Path(__file__).parent.parent.parent / "tests" / "data"

	filenames = [
		folder_data / "tables" / "generic.genotypes.3.xlsx",
		folder_data / "tables" / "generic.genotypes.5.xlsx",
		folder_data / "tables" / "generic.genotypes.10.xlsx",
		# folder_data / "tables" / "real.RMS1-G02.xlsx",
		folder_data / "tables" / "real.nature12344-s2.BYB1-G07.xlsx",
		folder_data / "tables" / "real.B1_muller_try1.xlsx",
		folder_data / "tables_input_trajectories" / "B1_Muller.xlsx",
		folder_data / "tables_input_trajectories" / "P1_Final_Muller.csv",
		folder_data / "tables_input_trajectories" / "P3_Muller.csv"
	]

	return filenames


def benchmark_serial(datasets: List[pandas.DataFrame], threads: int = 1):
	"""
		Benchmarks the serial (single-threaded) distance calculator.
	"""

	calculator = DistanceCalculator(0.0, 0.97, 'binomial', threads = threads)
	benchmarktable = list()
	for dataset in datasets:
		total_elements = len(dataset)
		total_timepoints = len(widgets.get_numeric_columns(dataset.columns))
		start = time.time()
		calculator.run(dataset)
		duration = time.time() - start

		logger.info(f"It took {duration:.2f} seconds to compute {total_elements} elements and {total_timepoints} timepoints.")
		row = {
			'totalTrajectories': total_elements,
			'totalTimepoints':   total_timepoints,
			'duration':          duration,
			'threads':           threads
		}
		benchmarktable.append(row)

	return pandas.DataFrame(benchmarktable)


def plot_banchmarks(data: pandas.DataFrame):
	fig, ax = plt.subplots(figsize = (12, 10))

	ax.scatter("totalTrajectories", "duration", data = data)
	plt.show()


def main():
	filenames_datasets = get_datasets()

	tables_datasets = load_datasets(filenames_datasets)
	#tables_datasets += load_highresolution_datasets()
	dfs = list()
	for i in range(1, 8):
		df_serial = benchmark_serial(tables_datasets, threads = i)
		dfs.append(df_serial)
	d = pandas.concat(dfs)
	d.to_csv("benchmarktable.tsv", sep = "\t")
	plot_banchmarks(dfs[1])


if __name__ == "__main__":
	main()



"""
	Implements methods that are not part of the scripts themselves, but offer a few utility methods for things such as converting the output
	of breseq into a trajectory table or benchmarking the distance calculators.
"""